# **************************************************************************
# *
# * Authors:     Pablo Conesa(pconesa@cnb.csic.es)
# *              Roberto Marabini(roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.protocol.params as params
from pwem.protocols import EMProtocol
from pwem.objects.data import SetOfParticles

from pwem.convert.transformations import (
    rotation_matrix, angle_between_vectors,
    vector_product)
import numpy as np

from pwem.constants import (SYM_I222, SYM_I222r, SYM_In25, SYM_In25r,
                            SYM_I2n3, SYM_I2n3r, SYM_I2n5, SYM_I2n5r,
                            SYM_DIHEDRAL_X, SYM_DIHEDRAL_Y,
                            SYM_TETRAHEDRAL_222, SYM_TETRAHEDRAL_Z3,
                            SYM_TETRAHEDRAL_Z3R, SCIPION_SYM_NAME
                            )


class ProtProjectionEditor(EMProtocol):
    """
    This protocol edits the projection directions of all the items of a set of particles using
    a formula. This could be useful for applying geometrical transformation to a set of
    particles.

    Several predefined operation are offered such as
    * apply the rotation matrix define by a origin vector and a target vector
    * rotate the projection vector around a given vector by  A degrees
    * convert between icosahedral symmetries
    * convert between dihedral symmetries
    * convert between tetrahedral symmetries
    """
    _label = 'edit projection'
    CHOICE_MOVE_VECTOR = 0
    CHOICE_ROTATE_VECTOR = 1
    CHOICE_ICOSAHEDRAL = 2
    CHOICE_DIHEDRAL = 3
    CHOICE_TETRAHEDRAL = 4
    CHOICE_LABEL = {
        CHOICE_MOVE_VECTOR: 'move source vector to target vector',
        CHOICE_ROTATE_VECTOR: 'rotate around vector',
        CHOICE_ICOSAHEDRAL: 'convert between icosahedral symmetries',
        CHOICE_DIHEDRAL: 'convert between dihedral symmetries',
        CHOICE_TETRAHEDRAL: 'convert between tetrahedral symmetries',
        }
    ICOSAHEDRAL_SYM_NAME = {
        SYM_I222: SCIPION_SYM_NAME[SYM_I222],
        SYM_I222r: SCIPION_SYM_NAME[SYM_I222r],
        SYM_In25: SCIPION_SYM_NAME[SYM_In25],
        SYM_In25r: SCIPION_SYM_NAME[SYM_In25r],
        SYM_I2n3: SCIPION_SYM_NAME[SYM_I2n3],
        SYM_I2n3r: SCIPION_SYM_NAME[SYM_I2n3r],
        SYM_I2n5: SCIPION_SYM_NAME[SYM_I2n5],
        SYM_I2n5r: SCIPION_SYM_NAME[SYM_I2n5r]
        }

    DIHEDRAL_SYM_NAME = {
        SYM_DIHEDRAL_X: SCIPION_SYM_NAME[SYM_DIHEDRAL_X],
        SYM_DIHEDRAL_Y: SCIPION_SYM_NAME[SYM_DIHEDRAL_Y]
        }

    TETRAHEDRAL_SYM_NAME = {
        SYM_TETRAHEDRAL_222: SCIPION_SYM_NAME[SYM_TETRAHEDRAL_222],
        SYM_TETRAHEDRAL_Z3: SCIPION_SYM_NAME[SYM_TETRAHEDRAL_Z3],
        SYM_TETRAHEDRAL_Z3R: SCIPION_SYM_NAME[SYM_TETRAHEDRAL_Z3R]
        }

    def _defineParams(self, form):
        """
        Defines the parameters the protocol form will show and its behaviour
        :param form:
        """
        form.addSection(label='Input')
        form.addParam('inputSet', params.PointerParam,
                      pointerClass='SetOfParticles',
                      label='Set of particles to edit',
                      help='Set which items will be modified.'
                           'Must be a set of particles with transformations.')
        form.addParam(
            'operation', params.EnumParam,
            choices=[self.CHOICE_LABEL[self.CHOICE_MOVE_VECTOR],
                     self.CHOICE_LABEL[self.CHOICE_ROTATE_VECTOR],
                     self.CHOICE_LABEL[self.CHOICE_ICOSAHEDRAL],
                     self.CHOICE_LABEL[self.CHOICE_DIHEDRAL],
                     self.CHOICE_LABEL[self.CHOICE_TETRAHEDRAL],
                     ],
            default=self.CHOICE_ICOSAHEDRAL,
            label="Select operation",
            help="Select operation to be performed in the set.\n"
                 " *rotate to vector* modifies the alignment matrix "
                 " so a reconstruction made from the images produces a "
                 " rotated reconstruction\n"
                 " *convert between icosahedral symmetry* will modify "
                 " the alignment matrix so the reconstruction will have "
                 " the new symmetry provided\n"
                 " *convert between dihedral symmetry* will modify "
                 " the alignment matrix so the reconstruction will have "
                 " the new symmetry provided\n"
                 " *convert between tetrahedral symmetry* will modify "
                 " the alignment matrix so the reconstruction will have "
                 " the new symmetry provided\n"
            )
        # move vector
        line = form.addLine(
            'source vector',
            condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
            help='Vector will be rotated until overlaps target vector')
        line.addParam('xs', params.FloatParam, default=0,
                      condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
                      label="x",
                      help="X Coordinate ")
        line.addParam('ys', params.FloatParam, default=0,
                      condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
                      label="y",
                      help="Y dim ")
        line.addParam('zs', params.FloatParam, default=1,
                      condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
                      label="z",
                      help="Z dim ")

        line = form.addLine(
            'target vector',
            condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
            help='source vector will be rotated until '
                 'overlaps target vector')
        line.addParam('xt', params.FloatParam, default=1,
                      condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
                      label="x",
                      help="X Coordinate ")
        line.addParam(
            'yt', params.FloatParam, default=0,
            condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
            label="y",
            help="Y dim ")
        line.addParam('zt', params.FloatParam, default=0,
                      condition="operation==%d" % self.CHOICE_MOVE_VECTOR,
                      label="z",
                      help="Z dim ")
        # rotate vector
        line = form.addLine(
            'rotate around vector',
            condition="operation==%d" % self.CHOICE_ROTATE_VECTOR,
            help='rotate around this vector by "angle" angles')
        line.addParam('x', params.FloatParam, default=0,
                      condition="operation==%d" % self.CHOICE_ROTATE_VECTOR,
                      label="x",
                      help="X coordinate ")
        line.addParam('y', params.FloatParam, default=0,
                      condition="operation==%d" % self.CHOICE_ROTATE_VECTOR,
                      label="y",
                      help="Y coordinate ")
        line.addParam('z', params.FloatParam, default=1,
                      condition="operation==%d" % self.CHOICE_ROTATE_VECTOR,
                      label="z",
                      help="Z coordinate")
        line.addParam('angle', params.FloatParam, default=0,
                      condition="operation==%d" % self.CHOICE_ROTATE_VECTOR,
                      label="angle (degrees)",
                      help="rotation angle ")
        # Icosahedral symmetry
        line = form.addLine(
            'Icosahedral symmetry',
            condition="operation==%d" % self.CHOICE_ICOSAHEDRAL,
            help='convert between icosahedral symmetries')
        line.addParam(
            "originSymmetryGroupI", params.EnumParam,
            condition="operation==%d" % self.CHOICE_ICOSAHEDRAL,
            label="origin",
            help="Source symmetry. ",
            choices=list(self.ICOSAHEDRAL_SYM_NAME.values()),
            default=SYM_I222r - SYM_I222,
            )
        line.addParam(
            "targetSymmetryGroupI", params.EnumParam,
            condition="operation==%d" % self.CHOICE_ICOSAHEDRAL,
            label="target",
            help="Target symmetry. ",
            choices=list(self.ICOSAHEDRAL_SYM_NAME.values()),
            default=SYM_I222 - SYM_I222
            )
        # Dihedral symmetry
        line = form.addLine(
            'Dihedral symmetry',
            condition="operation==%d" % self.CHOICE_DIHEDRAL,
            help='convert between dihedral symmetries')
        line.addParam(
            "originSymmetryGroupD", params.EnumParam,
            condition="operation==%d" % self.CHOICE_DIHEDRAL,
            label="origin",
            help="Source symmetry. ",
            choices=list(self.DIHEDRAL_SYM_NAME.values()),
            default=SYM_DIHEDRAL_X - SYM_DIHEDRAL_X,
            )
        line.addParam(
            "targetSymmetryGroupD", params.EnumParam,
            condition="operation==%d" % self.CHOICE_DIHEDRAL,
            label="target",
            help="Target symmetry. ",
            choices=list(self.DIHEDRAL_SYM_NAME.values()),
            default=SYM_DIHEDRAL_Y - SYM_DIHEDRAL_X,
            )

        # Tetrahedral symmetry
        line = form.addLine(
            'Tetrahedral symmetry',
            condition="operation==%d" % self.CHOICE_TETRAHEDRAL,
            help='convert between dihedral symmetries'
            )
        line.addParam(
            "originSymmetryGroupT", params.EnumParam,
            condition="operation==%d" % self.CHOICE_TETRAHEDRAL,
            label="origin",
            help="Source symmetry. ",
            choices=list(self.TETRAHEDRAL_SYM_NAME.values()),
            default=SYM_TETRAHEDRAL_Z3 - SYM_TETRAHEDRAL_222,
            )
        line.addParam(
            "targetSymmetryGroupT", params.EnumParam,
            condition="operation==%d" % self.CHOICE_TETRAHEDRAL,
            label="target",
            help="Target symmetry. ",
            choices=list(self.TETRAHEDRAL_SYM_NAME.values()),
            default=SYM_TETRAHEDRAL_222 - SYM_TETRAHEDRAL_222,
            )

    def _insertAllSteps(self):
        operation = self.operation.get()
        if operation == self.CHOICE_MOVE_VECTOR:
            self._insertFunctionStep('rotateStep')
        elif operation == self.CHOICE_ROTATE_VECTOR:
            self._insertFunctionStep('rotateVectorStep')
        elif operation == self.CHOICE_ICOSAHEDRAL:
            self._insertFunctionStep('rotateIcosaStep')
        elif operation == self.CHOICE_DIHEDRAL:
            self._insertFunctionStep('rotateDiStep')
        elif operation == self.CHOICE_TETRAHEDRAL:
            self._insertFunctionStep('rotateTetraStep')

    def rotateTetraStep(self):
        """ Compute rotation matrix between tetrahedral symmetries
            and apply it to the projection directions
        """
        from pwem.convert.symmetry import Tetrahedral
        origSym = self.originSymmetryGroupT.get() + SYM_TETRAHEDRAL_222
        targetSym = self.targetSymmetryGroupT.get() + SYM_TETRAHEDRAL_222
        print("origSym", origSym)
        print("targetSym", targetSym)
        tetrahedral = Tetrahedral(sym=origSym)
        matrix = tetrahedral.coordinateSystemTransform(origSym, targetSym)

        # convert to numpy array
        # and add extra row
        matrix = np.array(matrix)
        matrix = np.append(matrix, [[0, 0, 0, 1]], axis=0)

        inputSet = self.inputSet.get()
        modifiedSet = inputSet.createCopy(self._getExtraPath(), copyInfo=True)
        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            transformation = item.getTransform()
            transformation.composeTransform(matrix)
            modifiedSet.append(item)
        self.createOutput(self.inputSet, modifiedSet)

    def rotateDiStep(self):
        """ Compute rotation matrix from one dihedral
        symmetry to another and apply it to the projection directions
        """
        from pwem.convert.symmetry import Dihedral
        origSym = self.originSymmetryGroupD.get() + SYM_DIHEDRAL_X
        targetSym = self.targetSymmetryGroupD.get() + SYM_DIHEDRAL_X
        # n is a required parameter not used in this case
        dihedral = Dihedral(sym=origSym, n=1)
        matrix = dihedral.coordinateSystemTransform(origSym, targetSym)

        # convert to numpy array
        # and add extra row
        matrix = np.array(matrix)
        matrix = np.append(matrix, [[0, 0, 0, 1]], axis=0)

        inputSet = self.inputSet.get()
        modifiedSet = inputSet.createCopy(self._getExtraPath(), copyInfo=True)
        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            transformation = item.getTransform()
            transformation.composeTransform(matrix)
            modifiedSet.append(item)
        self.createOutput(self.inputSet, modifiedSet)

    def rotateIcosaStep(self):
        """
        Compute rotation matrix from one icosahedral
        symmetry to another and apply it
        to projection directions. Reconstructed volume
        should change symmetry
        """
        from pwem.convert.symmetry import Icosahedron
        origSym = self.originSymmetryGroupI.get() + SYM_I222
        targetSym = self.targetSymmetryGroupI.get() + SYM_I222
        ico = Icosahedron(orientation=SCIPION_SYM_NAME[origSym][1:])
        matrix = ico.coordinateSystemTransform(SCIPION_SYM_NAME[origSym][1:],
                                               SCIPION_SYM_NAME[targetSym][1:])
        # convert to numpy array
        # and add extra row
        matrix = np.array(matrix)
        matrix = np.append(matrix, [[0, 0, 0, 1]], axis=0)

        inputSet = self.inputSet.get()
        modifiedSet = inputSet.createCopy(self._getExtraPath(), copyInfo=True)
        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            transformation = item.getTransform()
            transformation.composeTransform(matrix)
            modifiedSet.append(item)
        self.createOutput(self.inputSet, modifiedSet)

    def createOutput(self, inputSet, modifiedSet):
        outputArgs = {inputSet.getExtended(): modifiedSet}
        self._defineOutputs(**outputArgs)

    def rotateStep(self):
        """
        Compute rotation matrix between user provided vectors
         (x,y,z) and (x',y',z') and apply it
        to projection directions. Reconstructed volume
        should rotate
        """
        v_source = np.array([self.xs.get(), self.ys.get(), self.zs.get()])
        v_source = v_source / np.linalg.norm(v_source)
        v_target = np.array([self.xt.get(), self.yt.get(), self.zt.get()])
        v_target = v_target / np.linalg.norm(v_target)
        matrix = rotation_matrix(angle_between_vectors(v_source, v_target),
                                 vector_product(v_source, v_target))
        print("rotateStep:matrix", matrix)
        inputSet = self.inputSet.get()
        modifiedSet = inputSet.createCopy(self._getExtraPath(), copyInfo=True)
        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            transformation = item.getTransform()
            transformation.composeTransform(matrix)
            modifiedSet.append(item)

        self.createOutput(self.inputSet, modifiedSet)

    def rotateVectorStep(self):
        """
        Compute rotation matrix around user provided vector
         (x,y,z) and rotate "angle" degrees the projection directions.
          Reconstructed volume should rotate around axis
        """
        v_source = np.array([self.x.get(), self.y.get(), self.z.get()])
        v_source = v_source / np.linalg.norm(v_source)
        angle = np.radians(self.angle.get())
        matrix = rotation_matrix(angle, v_source)
        print("matrix_rot_vector", matrix)
        inputSet = self.inputSet.get()
        modifiedSet = inputSet.createCopy(self._getExtraPath(), copyInfo=True)
        for sourceItem in inputSet.iterItems():
            item = sourceItem.clone()
            transformation = item.getTransform()
            transformation.composeTransform(matrix)
            modifiedSet.append(item)

        self.createOutput(self.inputSet, modifiedSet)

    def _validate(self):
        errors = []
        inputSet = self.inputSet.get()
        if not isinstance(inputSet, SetOfParticles):
            errors.append("The input data set is not a set of particles")
        if not inputSet.hasAlignmentProj():
            errors.append("The input data set does not have alignment 3D")
        return errors



# **************************************************************************
# *
# * Authors:     J.M. de la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import numpy
from PIL import Image
from tifffile import TiffFile

import pyworkflow.utils as pwutils
import pwem.objects as emobj
import pwem.constants as emcts
from .. import lib

import logging
logger = logging.getLogger(__name__)

class ImageHandler(object):
    """ Class to provide several Image manipulation utilities. """

    # TODO: remove dependency from Xmipp

    def __init__(self):
        # Now it will use Xmipp image library
        # to read and write most of formats, in the future
        # if we want to be independent of Xmipp, we should have
        # our own image library
        self._img = lib.Image()
        self._imgClass = lib.Image

    @classmethod
    def fixXmippVolumeFileName(cls, image):
        """ This method will add :mrc to .mrc volumes
        because for mrc format is not possible to distinguish
        between 3D volumes and 2D stacks.
        """
        # We can not import Volume from top level since
        # data depends on this module
        fn = image.getFileName()
        if isinstance(image, emobj.Volume):
            if fn.endswith('.mrc') or fn.endswith('.map'):
                fn += ':mrc'
        elif isinstance(image, emobj.Movie):
            if fn.endswith('.mrc'):
                fn += ':mrcs'
            elif fn.endswith('.em'):
                fn += ':ems'

        return fn

    @classmethod
    def locationToXmipp(cls, location):
        """ Convert an index and filename location
        to a string with @ as expected in Xmipp.
        """
        index, filename = cls._convertToLocation(location)
        if index != emcts.NO_INDEX:
            return "%06d@%s" % (index, filename)
        return filename

    @classmethod
    def _convertToLocation(cls, location):
        """ Get a location in a tuple format (index, filename).
        location could be:
            tuple -> (index, filename)
            string -> (NO_INDEX, filename)
            image -> (image.getIndex(), image.getFileName())
        """
        if isinstance(location, tuple):
            outLocation = location

        elif isinstance(location, str):
            outLocation = (emcts.NO_INDEX, location)

        elif hasattr(location, 'getLocation'):
            # This case includes Image and its subclasses
            outLocation = (location.getIndex(),
                           cls.fixXmippVolumeFileName(location))
        else:
            raise Exception('Can not convert object %s to (index, location)'
                            % type(location))

        return outLocation

    @classmethod
    def existsLocation(cls, locationObj):
        """ Return True if a given location exists. 
        Location have the same meaning than in _convertToLocation.
        """
        if locationObj is None:
            fn = None
        elif isinstance(locationObj, tuple):
            fn = locationObj[1]
        elif isinstance(locationObj, str):
            fn = locationObj
        elif hasattr(locationObj, 'getLocation'):
            # This case includes Image and its subclasses
            fn = locationObj.getLocation()[1]
        else:
            raise Exception('Can not match object %s to '
                            '(index, location)' % type(locationObj))

        # If either the location is None or location
        if fn is None:
            return False

        # Remove filename format specification such as :mrc, :mrcs or :ems
        if ':' in fn:
            fn = fn.split(':')[0]

        return os.path.exists(fn)

    @classmethod
    def getSupportedDataType(cls, inDataType, outputFilename):
        """ Returns the most similar data type supported by the
        output format"""
        outDataType = inDataType

        if outputFilename.endswith(".mrc") or outputFilename.endswith(".mrcs"):
            if inDataType == lib.DT_SCHAR:
                outDataType = lib.DT_USHORT

        return outDataType

    def convert(self, inputObj, outputObj, dataType=None, transform=None):
        """ Convert from one image to another.
        inputObj and outputObj can be: tuple, string, or Image subclass 
        (see self._convertToLocation)
        transform: if not None, apply this transformation
        """
        inputLoc = self._convertToLocation(inputObj)
        outputLoc = self._convertToLocation(outputObj)

        if outputLoc[1].lower().endswith('.img'):
            # FIXME Since now we can not read dm4 format in Scipion natively
            # we are opening an Eman2 process to read the dm4 file
            from pwem import Domain
            convertImage = Domain.importFromPlugin('eman2.convert',
                                                   'convertImage',
                                                   doRaise=True)
            convertImage(inputLoc, outputLoc)
        else:
            # Read from input
            self._img.read(inputLoc)

            if dataType is not None:
                self._img.convert2DataType(dataType)
            if transform is not None:
                self._img.applyTransforMatScipion(transform.getMatrixAsList())
            # Write to output
            self._img.write(outputLoc)

    def convertStack(self, inputFn, outputFn, firstImg=None, lastImg=None,
                     inFormat=None, outFormat=None):
        """ Convert an input stack file into another.
        It is possible to only use a subset of frames to be written in the
            output stack.
        If outFormat/inFomat=None then there will be
        inferred from extension.If firstFrame/lastFrame are not None, the output
        stack will be a subset of input stack. If it are none, the conversion is
        over the whole stack. If the input format is ".dm4" or  ".img" only is
        allowed the conversion of the whole stack.
        """
        # inputLower = inputFn.lower()
        outputLower = outputFn.lower()
        if outputLower.endswith('.img'):
            if (firstImg and lastImg) is None:
                # FIXME Since now we can not read dm4 format in Scipion natively
                # or writing recent .img format
                # we are opening an Eman2 process to read the dm4 file
                from pwem import Domain
                convertImage = Domain.importFromPlugin('eman2.convert',
                                                       'convertImage')
                convertImage(inputFn, outputFn)
            else:
                ext = os.path.splitext(outputFn)[1]
                raise Exception("if convert from %s, firstImg and lastImg "
                                "must be None" % ext)
        # elif inputLower.endswith('.tif'):
        #     # FIXME: It seems that we have some flip problem with compressed
        #     # tif files, we need to check that
        #     if outputLower.endswith('.mrc'):
        #         self.runJob('tif2mrc', '%s %s' % (inputFn, outputFn))
        #     else:
        #         raise Exception("Conversion from tif to %s is not "
        #                        "implemented yet. " % pwutils.getExt(outputFn))
        else:
            # get input dim
            (x, y, z, n) = lib.getImageSize(inputFn)

            location = self._convertToLocation(inputFn)
            self._img.read(location, lib.HEADER)
            dataType = self.getSupportedDataType(self._img.getDataType(),
                                                 outputLower)

            if (firstImg and lastImg) is None:
                n = max(z, n)
                firstImg = 1
                lastImg = n
            else:
                n = lastImg - firstImg + 1

            # Create empty output stack file to reserve desired space
            lib.createEmptyFile(outputFn, x, y, 1, n, dataType)
            for i, j in zip(range(firstImg, lastImg + 1), range(1, n + 1)):
                self.convert((i, inputFn), (j, outputFn))

    def getDimensions(self, locationObj):
        """ It will return a tuple with the images dimensions.
        The tuple will contains:
            (x, y, z, n) where x, y, z are image dimensions (z=1 for 2D) and 
            n is the number of elements if is a stack.
        """
        if self.existsLocation(locationObj):
            location = self._convertToLocation(locationObj)
            fn = location[1]

            # Dimensions based on Readers. Registered and defined in the bottom
            reader = ImageReadersRegistry.getReader(fn)
            return reader.getDimensions(fn)
            # ext = pwutils.getExt(fn).lower()
            #
            # # Local import to avoid import loop between ImageHandler and Ccp4Header.
            # from pwem.convert import headers
            #
            # if ext == '.png' or ext == '.jpg':
            #     im = Image.open(fn)
            #     x, y = im.size  # (width,height) tuple
            #     return x, y, 1, 1
            #
            # elif headers.getFileFormat(fn) == headers.MRC:
            #     header = headers.Ccp4Header(fn, readHeader=True)
            #     return header.getXYZN()
            #
            # elif ext == '.img':
            #     # FIXME Since now we can not read dm4 format in Scipion natively
            #     # or recent .img format
            #     # we are opening an Eman2 process to read the dm4 file
            #     from pwem import Domain
            #     getImageDimensions = Domain.importFromPlugin(
            #         'eman2.convert', 'getImageDimensions',
            #         doRaise=True)
            #     return getImageDimensions(fn)  # we are ignoring index here
            # elif ext in ['.eer', '.gain']:
            #     tif = TiffFile(fn)
            #     frames = len(tif.pages)  # number of pages in the file
            #     page = tif.pages[0]  # get shape and dtype of the image in the first page
            #     x, y = page.shape
            #     return x, y, frames, 1
            # else:
            #     self._img.read(location, lib.HEADER)
            #     return self._img.getDimensions()
        else:
            return None, None, None, None

    def getDataType(self, locationObj):
        if self.existsLocation(locationObj):
            location = self._convertToLocation(locationObj)
            self._img.read(location, lib.HEADER)
            return self._img.getDataType()
        else:
            return None

    def read(self, inputObj):
        """ Create a new Image class from inputObj 
        (inputObj can be tuple, str or Image subclass). """
        location = self._convertToLocation(inputObj)

        return self._imgClass(location)

    def createImage(self):
        return self._imgClass()

    def write(self, image, outputObj):
        """ Write to disk an image from outputObj 
        (outputObj can be tuple, str or Image subclass). """
        location = self._convertToLocation(outputObj)
        image.write(location)

    def compareData(self, locationObj1, locationObj2, tolerance=0.0001):
        """ Compare if two locations have the same binary data.
        """
        loc1 = self._convertToLocation(locationObj1)
        loc2 = self._convertToLocation(locationObj2)

        return lib.compareTwoImageTolerance(loc1, loc2, tolerance)

    def computeAverage(self, inputSet):
        """ Compute the average image either from filename or set.
        If inputSet is a filename, we will read the whole stack
        and compute the average from all images.
        If inputSet is a SetOfImages subclass, we will iterate
        and compute the average from all images.
        """
        if isinstance(inputSet, str):
            _, _, _, n = self.getDimensions(inputSet)
            if n:
                avgImage = self.read((1, inputSet))

                for i in range(2, n + 1):
                    self._img.read((i, inputSet))
                    avgImage.inplaceAdd(self._img)

                avgImage.inplaceDivide(n)
                return avgImage
        else:
            n = inputSet.getSize()
            if n:
                imageIter = iter(inputSet)
                img = next(imageIter)
                avgImage = self.read(img)

                for img in imageIter:
                    self._img.read(self._convertToLocation(img))
                    avgImage.inplaceAdd(self._img)

                avgImage.inplaceDivide(n)
                return avgImage

        return None

    def invertStack(self, inputFn, outputFn):
        # get input dim
        (x, y, z, n) = lib.getImageSize(inputFn)
        # Create empty output stack for efficiency
        lib.createEmptyFile(outputFn, x, y, z, n)
        # handle image formats
        for i in range(1, n + 1):
            self.invert((i, inputFn), (i, outputFn))

    def invert(self, inputObj, outputObj):
        """ invert the pixels.
        inputObj and outputObj can be: tuple, string, or Image subclass 
        (see self._convertToLocation)
        """
        # Read from input
        self._img.read(self._convertToLocation(inputObj))
        self._img.inplaceMultiply(-1)
        # Write to output
        self._img.write(self._convertToLocation(outputObj))

    @classmethod
    def __runXmippProgram(cls, program, args):
        """ Internal shortcut function to launch a Xmipp program. """
        from pwem import Domain
        xmipp3 = Domain.importFromPlugin('xmipp3')
        xmipp3.Plugin.runXmippProgram(program, args)

    @classmethod
    def __runEman2Program(cls, program, args):
        """ Internal workaround to launch an EMAN2 program. """
        from pwem import Domain
        eman2 = Domain.importFromPlugin('eman2', doRaise=True)
        from pyworkflow.utils.process import runJob
        runJob(None, eman2.Plugin.getProgram(program), args,
               env=eman2.Plugin.getEnviron())

    def createCircularMask(self, radius, refImage, outputFile):
        """ Create a circular mask with the given radius (pixels)
        and with the same dimensions of the refImage.
        The radius should be less or equal dim(refImage)/2
        The mask will be stored in 'outputFile'
        """
        inputRef = self.locationToXmipp(refImage)
        self.__runXmippProgram('xmipp_transform_mask',
                               '-i %s --create_mask  %s --mask circular -%d'
                               % (inputRef, outputFile, radius))

    def rotateVolume(self, inputFile, outputFile, transformation):
        """ Apply geometric transformations to images. You can shift, rotate
        and scale a group of images/volumes.
        """
        elementList = [str(item) for item in transformation.getRotationMatrix().flatten().tolist()]
        unrolledMatrix = ' '.join(elementList)
        self.__runXmippProgram('xmipp_transform_geometry',
                               '-i %s -o %s --rotate_volume matrix %s' % (inputFile, outputFile, unrolledMatrix))

    def addNoise(self, inputFile, outputFile, std=1., avg=0.):
        """ Add Gaussian noise to an input image (or stack)
        to produce noisy images.
        Params:
            inputFile: the filename of the input images
            outputFile: the filename of the output noisy images
            noiseStd: standard deviation for the Gaussian noise.
        """
        self.__runXmippProgram('xmipp_transform_add_noise',
                               '-i %s -o %s --type gaussian %f %f'
                               % (inputFile, outputFile, std, avg))

    def truncateMask(self, inputFile, outputFile, newDim=None):
        """ Forces the values of a mask to be between 0 and 1.
        Additionally, the output mask can be scaled to a new dimension.

        Params:
            inputFile: the filename of the input either image or volume
            outputFile: the filename of the output either image or volume
            newDim: scale the output Mask to a new dimension if not None
        """
        if inputFile.endswith('.mrc'):
            inputFile += ':mrc'

        ioStr = '-i %s -o %s' % (inputFile, outputFile)

        if newDim:
            self.__runXmippProgram('xmipp_image_resize',
                                   '%s --dim %d' % (ioStr, newDim))
            ioStr = '-i %s' % outputFile
        self.__runXmippProgram('xmipp_transform_threshold',
                               '%s --select below 0 --substitute '
                               'value 0' % ioStr)

        self.__runXmippProgram('xmipp_transform_threshold',
                               '-i %s --select above 1 --substitute '
                               'value 1' % outputFile)

    @classmethod
    def createEmptyImage(cls, fnOut, xDim=1, yDim=1, zDim=1, nDim=1,
                         dataType=None):
        dt = dataType or lib.DT_FLOAT
        lib.createEmptyFile(fnOut, xDim, yDim, zDim, nDim, dt)

    @classmethod
    def isImageFile(cls, imgFn):
        """ Check if imgFn has an image extension. The function
        is implemented in the Xmipp binding.
        """
        # Local import to avoid import loop between ImageHandler and Ccp4Header.
        from pwem.convert import headers

        return (lib.FileName(imgFn).isImage() or
                headers.getFileFormat(imgFn) == headers.MRC)

    def computeThumbnail(self, inputFn, outputFn, scaleFactor=6, flipOnY=False,
                         flipOnX=False):
        """ Compute a thumbnail of inputFn, save to ouptutFn.
        Optionally choose a scale factor eg scaleFactor=6 will make
        a thumbnail 6 times smaller.
        """
        outputFn = outputFn or self.getThumbnailFn(inputFn)
        args = '"%s" "%s" ' % (inputFn, outputFn)

        process = "--process normalize"
        process += '' if not flipOnY else " --process=xform.flip:axis=y"
        process += '' if not flipOnX else " --process=xform.flip:axis=x"

        args += "--fouriershrink %s %s" % (scaleFactor, process)

        self.__runEman2Program('e2proc2d.py', args)

        return outputFn

    @staticmethod
    def getThumbnailFn(inputFn):
        """Replace the extension in inputFn with thumb.png"""
        return pwutils.replaceExt(inputFn, "thumb.png")

    @classmethod
    def getVolFileName(cls, location):
        if isinstance(location, tuple):
            fn = location[1]
        elif isinstance(location, str):
            fn = location
        elif hasattr(location, 'getLocation'):
            fn = location.getLocation()[1]
        else:
            raise Exception('Can not match object %s to (index, location)'
                            % type(location))

        if fn.endswith('.mrc') or fn.endswith('.map'):
            fn += ':mrc'

        return fn

    @classmethod
    def removeFileType(cls, fileName):
        # Remove filename format specification such as :mrc, :mrcs or :ems
        if ':' in fileName:
            fileName = fileName.split(':')[0]
        return fileName

    @classmethod
    def scaleFourier(cls, inputFn, outputFn, scaleFactor):
        """ Scale an image by cropping in Fourier space. """
        # TODO: Avoid using xmipp program for this
        cls.__runXmippProgram("xmipp_transform_downsample",
                              "-i %s -o %s --step %f --method fourier"
                              % (inputFn, outputFn, scaleFactor))

    @classmethod
    def scaleSplines(cls, inputFn, outputFn, scaleFactor, finalDimension=None,
                     forceVolume=False):
        """ Scale an image using splines. """
        I = lib.Image()
        I.read(inputFn)
        x, y, z, n = I.getDimensions()
        setDimensions = 0
        if forceVolume:
            z = max(z, n)
            setDimensions = 1
        if finalDimension is not None:
            x, y, z = int(finalDimension), int(finalDimension), int(finalDimension)
        else:
            x, y, z = int(x * scaleFactor), int(y * scaleFactor), int(z * scaleFactor)

        I.scale(x, y, z, setDimensions)
        I.write(outputFn)

    @classmethod
    def scale2DStack(cls, inputFn, outputFn, scaleFactor=None, finalDimension=None):
        """
         Scale a 2D images stack using PIL.
        """
        from PIL import Image
        if scaleFactor is None and finalDimension is None:
            raise TypeError("scaleFactor or finalDimension must be passed")

        I = lib.Image()
        I.read(inputFn)
        x, y, z, n = I.getDimensions()

        if not finalDimension:
            finalDimension = round(x*scaleFactor)

        objects = I.getData()

        newStack = numpy.zeros((n, 1, finalDimension, finalDimension))
        count = 0
        for object in objects:
            slice = object[0]
            image = Image.fromarray(slice)
            resizedImage = image.resize((finalDimension, finalDimension))
            newStack[count] = numpy.array(resizedImage)
            count += 1

        I.setData(newStack)
        I.write(outputFn)

    @staticmethod
    def applyTransform(inputFile, outputFile, transformMatrix, shape, fillValue=None, doWrap=False):
        """Apply the transformation matrix over the input image and return the transformed image in a given shape.
        Parameters:
            :param inputFile: path location of the input image
            :param outputFile: path location of the output image
            :param transformMatrix: transformation matrix to be applied to the image. It should be a numpy array data type.
            :param shape: dimensions of the output image given as a tuple (yDim, xDim)
            :param fillValue: value with which the empty regions due to wrapping will be filled.
            :param doWrap: if True the empty regions will be filled with the wrapped values of the input image.
        """
        imageObj = lib.Image()
        imageObj.read(inputFile)
        if fillValue is None:
            mean, _, _, _ = imageObj.computeStats()
            resultImage = imageObj.applyWarpAffine(list(transformMatrix.flatten()), shape, doWrap, mean)
        else:
            resultImage = imageObj.applyWarpAffine(list(transformMatrix.flatten()), shape, doWrap, fillValue)
        resultImage.write(outputFile)


class ImageReader:
    @staticmethod
    def getCompatibleExtensions()-> list:
        """ Returns a list of the compatible extensions the reader can handle"""
        pass
    @staticmethod
    def getDimensions(filePath):
        """ Returns the dimensions [X,Y,Z,N] of the file"""
        pass

class PILImageReader(ImageReader):
    """ PIL image reader"""
    @staticmethod
    def getCompatibleExtensions() -> list:
        return ['png','jpg', 'jpeg']
    @staticmethod
    def getDimensions(filePath):
        im = Image.open(filePath)
        x, y = im.size  # (width,height) tuple
        return x, y, 1, 1

class TiffImageReader(ImageReader):
    """ Tiff image reader"""
    @staticmethod
    def getCompatibleExtensions() -> list:
        return ['tif','tiff', 'gain', 'eer']

    @staticmethod
    def getDimensions(filePath):

        tif = TiffFile(filePath)
        frames = len(tif.pages)  # number of pages in the file
        page = tif.pages[0]  # get shape and dtype of the image in the first page

        # TO confirm but dimensions seems to be swap or at least in synx with page.axes
        # So we are using axes attribute to get x and y
        x, y = page.shape # IMPORTANT: to match xmipp convention
        axes = page.axes.lower()
        if axes.startswith("y"):
            x,y = y,x

        return x, y, 1, frames

class EMANImageReader(ImageReader):
    """ Image reader for eman file formats"""
    @staticmethod
    def getCompatibleExtensions() -> list:
        return ["img"]

    @staticmethod
    def getDimensions(filePath):

        from pwem import Domain
        getImageDimensions = Domain.importFromPlugin(
            'eman2.convert', 'getImageDimensions',
            doRaise=True)
        return getImageDimensions(filePath)  # we are ignoring index here

class XMIPPImageReader(ImageReader):
    @staticmethod
    def getCompatibleExtensions():
        return emcts.ALL_MRC_EXTENSIONS + emcts.ALL_TIF_EXTENSIONS + ["hdf5", "dm4", "stk", "spi", "vol", "tif"]

    @staticmethod
    def getDimensions(filePath):
        img = lib.Image()
        img.read(filePath, lib.HEADER)
        return img.getDimensions()

class MRCImageReader(ImageReader):
    """ Image reader for MRC files"""
    @staticmethod
    def getCompatibleExtensions() -> list:
        return emcts.ALL_MRC_EXTENSIONS

    @staticmethod
    def getDimensions(filePath):
        from pwem.convert import headers
        header = headers.Ccp4Header(filePath, readHeader=True)
        return header.getXYZN()

class ImageReadersRegistry:
    """ Class to register image readers to provide basic information about an image like dimensions or getting an image"""
    _readers = dict() # Dictionary to hold the readers. The key is the extension

    @classmethod
    def addReader(cls, imageReader:ImageReader):

        for ext in imageReader.getCompatibleExtensions():
            ext_low=ext.lower()
            logger.debug("Adding %s as image reader for %s" % (imageReader, ext_low))
            cls._readers[ext_low]= imageReader

    @classmethod
    def getReader(cls, filePath):
        """ Returns the reader or None able to deal with filePath based on the extension."""
        ext = filePath.split(".")[-1].lower()
        logger.debug("Getting ImageReader for %s (%s)" % (filePath, ext))

        reader = cls._readers.get(ext, None)

        # Fall back to XmippImage reader
        if reader is None:
            logger.info("Reader not registered for %s files. Falling back to XmippReader" % ext)
            reader = XMIPPImageReader

        return reader

# Register reader in the registry. Latest registered will take priority.
ImageReadersRegistry.addReader(XMIPPImageReader)
ImageReadersRegistry.addReader(MRCImageReader)
ImageReadersRegistry.addReader(EMANImageReader)
ImageReadersRegistry.addReader(PILImageReader)
ImageReadersRegistry.addReader(TiffImageReader)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import time

try:
    from itertools import izip
except ImportError:
    izip = zip

import pyworkflow.object as pwobj
import pyworkflow.protocol as pwprot
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

import pwem.objects as emobj
import pwem.constants as emcts


class EMProtocol(pwprot.Protocol):
    """ Base class to all EM protocols.
    It will contains some common functionalities. 
    """
    _base = True

    def __init__(self, **kwargs):
        pwprot.Protocol.__init__(self, **kwargs)

    def __createSet(self, SetClass, template, suffix, **kwargs):
        """ Create a set and set the filename using the suffix.
        If the file exists, it will be delete. """
        setFn = self._getPath(template % suffix)
        # Close the connection to the database if
        # it is open before deleting the file
        pwutils.cleanPath(setFn)
        setObj = SetClass(filename=setFn, **kwargs)
        return setObj

    def _createSetOfMicrographs(self, suffix=''):
        return self.__createSet(emobj.SetOfMicrographs,
                                'micrographs%s.sqlite', suffix,
                                indexes=['_index'])

    def _createSetOfCoordinates(self, micSet, suffix=''):
        coordSet = self.__createSet(emobj.SetOfCoordinates,
                                    'coordinates%s.sqlite', suffix,
                                    indexes=['_micId'])
        coordSet.setMicrographs(micSet)
        return coordSet

    def _createCoordinatesTiltPair(self, micTiltPairs, uCoords, tCoords,
                                   angles, suffix):
        coordTiltPairs = self.__createSet(emobj.CoordinatesTiltPair,
                                          'coordinates_pairs%s.sqlite', suffix,
                                          indexes=['_untilted._micId',
                                                   '_tilted._micId'])
        coordTiltPairs.setUntilted(uCoords)
        coordTiltPairs.setTilted(tCoords)
        coordTiltPairs.setAngles(angles)
        coordTiltPairs.setMicsPair(micTiltPairs)

        for coordU, coordT in izip(uCoords, tCoords):
            coordTiltPairs.append(emobj.TiltPair(coordU, coordT))

        return coordTiltPairs

    def _createSetFromName(self, className, suffix=''):
        """ Create a new set from the given className. """
        _createSetFunc = getattr(self, '_create%s' % className)
        return _createSetFunc(suffix)

    def _createSetOfImages(self, suffix=''):
        return self.__createSet(emobj.SetOfImages, 'images%s.sqlite', suffix)

    def _createSetOfParticles(self, suffix=''):
        return self.__createSet(emobj.SetOfParticles, 'particles%s.sqlite', suffix,
                                indexes=['_classId', '_micId'])

    def _createSetOfAverages(self, suffix=''):
        return self.__createSet(emobj.SetOfAverages, 'averages%s.sqlite', suffix)

    def _createSetOfMovieParticles(self, suffix=''):
        return self.__createSet(emobj.SetOfMovieParticles,
                                'movie_particles%s.sqlite', suffix)

    def _createSetOfClasses2D(self, imgSet, suffix=''):
        classes = self.__createSet(emobj.SetOfClasses2D,
                                   'classes2D%s.sqlite', suffix)
        classes.setImages(imgSet)
        return classes

    def _createSetOfClasses3D(self, imgSet, suffix=''):
        classes = self.__createSet(emobj.SetOfClasses3D,
                                   'classes3D%s.sqlite', suffix)
        classes.setImages(imgSet)
        return classes

    def _createSetOfClassesVol(self, suffix=''):
        return self.__createSet(emobj.SetOfClassesVol, 'classesVol%s.sqlite', suffix)

    def _createSetOfVolumes(self, suffix=''):
        return self.__createSet(emobj.SetOfVolumes, 'volumes%s.sqlite', suffix)

    def _createSetOfCTF(self, suffix=''):
        return self.__createSet(emobj.SetOfCTF, 'ctfs%s.sqlite', suffix)

    def _createSetOfNormalModes(self, suffix=''):
        return self.__createSet(emobj.SetOfNormalModes, 'modes%s.sqlite', suffix)

    def _createSetOfMovies(self, suffix=''):
        return self.__createSet(emobj.SetOfMovies, 'movies%s.sqlite', suffix)

    def _createSetOfAngles(self, suffix=''):
        return self.__createSet(emobj.SetOfAngles,
                                'tiltpairs_angles%s.sqlite', suffix)

    def _createSetOfFSCs(self, suffix=''):
        return self.__createSet(emobj.SetOfFSCs, 'fscs%s.sqlite', suffix)

    def _createSetOfPDBs(self, suffix=''):
        return self.__createSet(emobj.SetOfAtomStructs, 'pdbs%s.sqlite', suffix)

    def _defineSourceRelation(self, srcObj, dstObj):
        """ Add a DATASOURCE relation between srcObj and dstObj """
        self._defineRelation(pwobj.RELATION_SOURCE, srcObj, dstObj)

    def _defineTransformRelation(self, srcObj, dstObj):
        self._defineRelation(pwobj.RELATION_TRANSFORM, srcObj, dstObj)
        # A transform relation allways implies a source relation
        self._defineSourceRelation(srcObj, dstObj)

    def _defineCtfRelation(self, micsObj, ctfsObj):
        self._defineRelation(emcts.RELATION_CTF, micsObj, ctfsObj)
        # A ctf relation allways implies a source relation
        self._defineSourceRelation(micsObj, ctfsObj)

    def _insertChild(self, key, child):
        if isinstance(child, emobj.Set):
            child.write()
        pwprot.Protocol._insertChild(self, key, child)

    def _validateDim(self, obj1, obj2, errors,
                     label1='Input 1', label2='Input 2'):
        """ Validate that obj1 and obj2 has the same dimensions.
        Params:
            obj1, obj2: input objects that can be Images or SetOfImages subclasses.
            errors: an error list where to append the error if not same dimensions
            label1, label2: labels for input objects used for the error message.
        """
        if obj1 is not None and obj2 is not None:
            d1 = obj1.getXDim()
            d2 = obj2.getXDim()

            if d1 is None:
                errors.append("Can not get dimensions from %s." % label1)
            elif d2 is None:
                errors.append("Can not get dimensions from %s." % label2)
            elif d1 != d2:
                msg = '%s and %s have not the same dimensions, \n' % (label1, label2)
                msg += 'which are %d and %d, respectively' % (d1, d2)
                errors.append(msg)

    def __str__(self):
        return self.getObjLabel()

    def allowsDelete(self, obj):
        if (isinstance(obj, emobj.SetOfCoordinates) or
                isinstance(obj, emobj.CoordinatesTiltPair)):
            return True
        return False

    # ------ Methods for Streaming picking --------------

    def _defineStreamingParams(self, form):
        """ This function can be called during the _defineParams method
        of some protocols that support stream processing.
        It will add an Streaming section together with the following
        params:
            streamingSleepOnWait: Some streaming protocols are quite fast,
                so, checking input/ouput updates creates an IO overhead.
                This params allows them to sleep (without consuming resources)
                to wait for new work to be done.
            streamingBatchSize: For some programs it is more efficient to process
                many items at once and not one by one. So this parameter will
                allow to group a number of items to be processed in the same
                protocol step. This can also reduce some IO overhead and spawning
                new OS processes.
        """
        form.addSection("Streaming")
        form.addParam("streamingWarning", params.LabelParam, important=True,
                      label="The following params are related to how "
                            "streaming is done in Scipion.")
        form.addParam("streamingSleepOnWait", params.IntParam, default=0,
                      label="Sleep when waiting (secs)",
                      help="If you specify a value greater than zero, "
                           "it will be the number of seconds that the "
                           "protocol will sleep when waiting for new "
                           "input data in streaming mode. ")
        form.addParam("streamingBatchSize", params.IntParam, default=1,
                      label="Batch size",
                      help="This value allows to group several items to be "
                           "processed inside the same protocol step. You can "
                           "use the following values: \n"
                           "*1*    The default behavior, the items will be "
                           "processed one by one.\n"
                           "*0*    Put in the same step all the items "
                           "available. If the sleep time is short, it could be "
                           "practically the same of one by one. If not, you "
                           "could have steps with more items. If the steps will "
                           "be executed in parallel, it is better not to use "
                           "this option.\n"
                           "*>1*   The number of items that will be grouped into "
                           "a step.")

    def _getStreamingSleepOnWait(self):
        return self.getAttributeValue('streamingSleepOnWait', 0)

    def _getStreamingBatchSize(self):
        return self.getAttributeValue('streamingBatchSize', 1)

    def _streamingSleepOnWait(self):
        """ This method should be used by protocols that want to sleep
        when there is not more work to do.
        """
        sleepOnWait = self._getStreamingSleepOnWait()
        if sleepOnWait > 0:
            self.info("Not much work to do now, sleeping %s seconds."
                      % sleepOnWait)
            time.sleep(sleepOnWait)

    def _insertNewMics(self, inputMics, getMicKeyFunc,
                       insertStepFunc, insertStepListFunc, *args):
        """ Insert steps of new micrographs taking into account the batch size.
        It is assumed that a self.micDict exists mapping between micKey and mic.
        It is also assumed that self.streamClosed is defined...with True value
        if the input stream is closed.
        This function can be used from several base protocols that support
        streaming and batch:

        - ProtCTFMicrographs
        - ProtParticlePickingAuto
        - ProtExtractParticles
        Params:
            inputMics: the input micrographs to be inserted into steps
            getMicKeyFunc: function to get the key of a micrograph
                (usually mic.getMicName()
            insertStepFunc: function used to insert a single step
            insertStepListFunc: function used to insert many steps.
            *args: argument list to be passed to step functions
        Returns:
            The list of step Ids that can be used as dependencies.
        """
        deps = []
        insertedMics = inputMics

        # Despite this function only should insert new micrographs
        # let's double check that they are not inserted already
        micList = [mic for mic in inputMics
                   if getMicKeyFunc(mic) not in self.micDict]

        def _insertSubset(micSubset):
            stepId = insertStepListFunc(micSubset, self.initialIds, *args)
            deps.append(stepId)

        # Now handle the steps depending on the streaming batch size
        batchSize = self._getStreamingBatchSize()

        if batchSize == 1:  # This is one by one, as before the batch size
            for mic in micList:
                stepId = insertStepFunc(mic, self.initialIds, *args)
                deps.append(stepId)
        elif batchSize == 0:  # Greedy, take all available ones
            _insertSubset(micList)
        else:  # batchSize > 0, insert only batches of this size
            n = len(inputMics)
            d = int(n / batchSize)  # number of batches to insert
            nd = d * batchSize
            for i in range(d):
                _insertSubset(micList[i * batchSize:(i + 1) * batchSize])

            if n > nd and self.streamClosed:  # insert last ones
                _insertSubset(micList[nd:])
            else:
                insertedMics = micList[:nd]

        for mic in insertedMics:
            self.micDict[getMicKeyFunc(mic)] = mic

        return deps



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import enum
import logging
import os.path

logger = logging.getLogger(__name__)
from os.path import exists, basename, abspath, relpath, join, splitext
from os import stat
from numpy import array
from numpy.linalg import norm
import glob

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params
import pwem.objects as emobj
import pwem.convert as emconv
from pwem import emlib

from .base import ProtImportFiles
from .images import ProtImportImages

class ImportVolumeOutputs(enum.Enum):
    outputVolume = emobj.Volume
    outputVolumes = emobj.SetOfVolumes

class ProtImportVolumes(ProtImportImages):
    """Protocol to import a set of volumes to the project"""
    _outputClassName = 'SetOfVolumes'
    _label = 'import volumes'
    _possibleOutputs = ImportVolumeOutputs
    IMPORT_FROM_EMDB = 1

    def __init__(self, **args):
        ProtImportImages.__init__(self, **args)

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2, relion...etc.
        """
        return ['local file', 'EMDBid']

    def _defineAcquisitionParams(self, form):
        """ Define acquisition parameters, it can be overwritten
        by subclasses to change what parameters to include.
        """
        param = form.getParam('importFrom')
        # Customize the help of this parameter with specific information
        # of the import particles
        param.help.set('Import 3D map data from EMDB server or local file')
        form.addParam('emdbId', params.IntParam,
                      condition='importFrom != IMPORT_FROM_FILES',
                      label="EMDB map ID (integer)", allowsNull=True,
                      help='Type an EMDB ID (Integer). ')
        form.addParam('setHalfMaps', params.BooleanParam,
                      condition='importFrom == IMPORT_FROM_FILES',
                      label='Set half maps',
                      help='Option YES:\nAssign two half maps to the imported map.',
                      default=False)
        form.addParam('half1map', params.PathParam,
                      label='Path half map1', help='Select first half map',
                      condition='setHalfMaps')
        form.addParam('half2map', params.PathParam,
                      label='Path half map2', help='Select second half map',
                      condition='setHalfMaps')
        form.addParam('samplingRate', params.FloatParam,
                      condition='importFrom == IMPORT_FROM_FILES',
                      label=pwutils.Message.LABEL_SAMP_RATE)
        form.addParam('setOrigCoord', params.BooleanParam,
                      label="Set origin of coordinates",
                      help="Option YES:\nA new volume will be created with "
                           "the "
                           "given ORIGIN of coordinates. This ORIGIN will be "
                           "set in the map file header.\nThe ORIGIN of "
                           "coordinates will be placed at the center of the "
                           "whole volume if you select n(x)/2, n(y)/2, "
                           "n(z)/2 as "
                           "x, y, z coordinates (n(x), n(y), n(z) are the "
                           "dimensions of the whole volume). However, "
                           "selecting "
                           "0, 0, 0 as x, y, z coordinates, the volume will be "
                           "placed at the upper right-hand corner.\n\n"
                           "Option NO:\nThe ORIGIN of coordinates will be "
                           "placed at the center of the whole volume ("
                           "coordinates n(x)/2, n(y)/2, n(z)/2 by default). "
                           "This "
                           "ORIGIN will NOT be set in the map file header.\n\n"
                           "WARNING: In case you want to process "
                           "the volume with programs requiring a specific "
                           "symmetry regarding the origin of coordinates, "
                           "for example the protocol extract unit "
                           "cell, check carefully that the coordinates of the "
                           "origin preserve the symmetry of the whole volume. "
                           "This is particularly relevant for loading "
                           "fragments/subunits of the whole volume.\n",
                      default=False)
        line = form.addLine('Offset',
                            help="A wizard will suggest you possible "
                                 "coordinates for the ORIGIN. In MRC volume "
                                 "files, the ORIGIN coordinates will be "
                                 "obtained from the file header.\n "
                                 "In case you prefer set your own ORIGIN "
                                 "coordinates, write them here. You have to "
                                 "provide the map center coordinates in "
                                 "Angstroms (pixels x sampling).\n",
                            condition='setOrigCoord')
        # line.addParam would produce a nicer looking form
        # but them the wizard icon is drawn outside the visible
        # window. Until this bug is fixed form is a better option
        form.addParam('x', params.FloatParam, condition='setOrigCoord',
                      label="x", help="offset along x axis (Angstroms)")
        form.addParam('y', params.FloatParam, condition='setOrigCoord',
                      label="y", help="offset along y axis (Angstroms)")
        form.addParam('z', params.FloatParam, condition='setOrigCoord',
                      label="z", help="offset along z axis (Angstroms)")

    def _insertAllSteps(self):
        if self.importFrom == self.IMPORT_FROM_FILES:
            self._insertFunctionStep(self.importFromFileStep,
                                 self.getPattern(),
                                 self.samplingRate.get(),
                                 self.setOrigCoord.get())
        else:
            self._insertFunctionStep(self.importFromEMDBStep)

    # --------------------------- STEPS functions -----------------------------

    def importFromFileStep(self, pattern, samplingRate, setOrigCoord=False):
        """ Copy images matching the filename pattern
        Register other parameters.
        """
        volSet = self._createSetOfVolumes()
        vol = emobj.Volume()

        self.info("Using pattern: '%s'" % pattern)

        # Create a Volume template object
        vol.setSamplingRate(samplingRate)

        imgh = emlib.image.ImageHandler()

        volSet.setSamplingRate(samplingRate)

        for fileName, fileId in self.iterFiles():
            x, y, z, n = imgh.getDimensions(fileName)
            if fileName.endswith('.mrc') or fileName.endswith('.map'):
                fileName += ':mrc'
                if z == 1 and n != 1:
                    zDim = n
                    n = 1
                else:
                    zDim = z
            else:
                zDim = z
            origin = emobj.Transform()
            if setOrigCoord:
                origin.setShiftsTuple(self._getOrigCoord())
            else:
                origin.setShifts(x / -2. * samplingRate,
                                 y / -2. * samplingRate,
                                 zDim / -2. * samplingRate)

            vol.setOrigin(origin)  # read origin from form

            if self.copyFiles or setOrigCoord:
                newFileName = abspath(self._getVolumeFileName(fileName, "mrc"))
                emconv.Ccp4Header.fixFile(fileName, newFileName, origin.getShifts(),
                                          samplingRate, emconv.Ccp4Header.ORIGIN)
                if self.setHalfMaps.get():
                    newFileName1 = abspath(self._getVolumeFileName(self.half1map.get(), "mrc"))
                    emconv.Ccp4Header.fixFile(self.half1map.get(), newFileName1, origin.getShifts(),
                                              samplingRate, emconv.Ccp4Header.ORIGIN)
                    newFileName2 = abspath(self._getVolumeFileName(self.half2map.get(), "mrc"))
                    emconv.Ccp4Header.fixFile(self.half2map.get(), newFileName2, origin.getShifts(),
                                              samplingRate, emconv.Ccp4Header.ORIGIN)

                    vol.setHalfMaps([relpath(newFileName1), relpath(newFileName2)])
            else:
                newFileName = abspath(self._getVolumeFileName(fileName))

                if fileName.endswith(':mrc'):
                    fileName = fileName[:-4]

                pwutils.createAbsLink(fileName, newFileName)
                if self.setHalfMaps.get():
                    pwutils.createAbsLink(self.half1map.get(),
                                          abspath(self._getVolumeFileName(self.half1map.get())))
                    pwutils.createAbsLink(self.half2map.get(),
                                          abspath(self._getVolumeFileName(self.half2map.get())))

                    vol.setHalfMaps([relpath(self._getVolumeFileName(self.half1map.get())),
                                     relpath(self._getVolumeFileName(self.half2map.get()))
                                     ])

            # Make newFileName relative
            # https://github.com/I2PC/scipion/issues/1935
            newFileName = relpath(newFileName)
            if n == 1:
                vol.cleanObjId()
                vol.setFileName(newFileName)
                vol.setObjComment(vol.getBaseName())
                volSet.append(vol)
            else:
                for index in range(1, n + 1):
                    vol.cleanObjId()
                    vol.setLocation(index, newFileName)
                    vol.setObjComment("%s@%s" % (index, vol.getBaseName()))
                    volSet.append(vol)

        if volSet.getSize() > 1:
            self._defineOutputs(**{ImportVolumeOutputs.outputVolumes.name:volSet})
        else:
            self._defineOutputs(**{ImportVolumeOutputs.outputVolume.name:vol})

    def importFromEMDBStep(self):
        """ Copy images matching the filename pattern
        Register other parameters.
        """
        vol = emobj.Volume()

        self.info("Downloading map with ID = %s" % self.emdbId)

        localFileName, sampling, origin = \
            fetch_emdb_map(self.emdbId,
                           self._getExtraPath(),
                           self._getTmpPath())

        # open volume and fill sampling and origin
        self.samplingRate.set(sampling)
        self._store(self.samplingRate)
        vol.setSamplingRate(sampling)
        vol.setFileName(localFileName)
        from pwem.objects.data import Transform
        originMat = Transform()
        originMat.setShifts(origin[0], origin[1], origin[2])
        vol.setOrigin(originMat)

        self._defineOutputs(**{ImportVolumeOutputs.outputVolume.name:vol})


    # --------------------------- INFO functions ------------------------------

    def _getVolMessage(self):
        if self.hasAttribute('outputVolume'):
            return "Volume %s" % self.getObjectTag('outputVolume')
        else:
            return "Volumes %s" % self.getObjectTag('outputVolumes')

    def _summary(self):
        summary = []
        if self.hasAttribute('outputVolume') or \
                self.hasAttribute('outputVolumes'):
            summary.append("%s imported from:\n%s" % (self._getVolMessage(),
                                                      self.getPattern()))

            summary.append(u"Sampling rate: *%0.2f* (Å/px)" %
                           self.samplingRate.get())
        return summary

    def _methods(self):
        methods = []
        if self.hasAttribute('outputVolume') or \
                self.hasAttribute('outputVolumes'):
            methods.append(" %s imported with a sampling rate *%0.2f*" %
                           (self._getVolMessage(), self.samplingRate.get()), )
        return methods

    def _getVolumeFileName(self, fileName, extension=None):
        if extension is not None:
            baseFileName = "import_" + basename(fileName).split(".")[0] + ".%s" % extension
        else:
            baseFileName = "import_" + basename(fileName).split(":")[0]

        return self._getExtraPath(baseFileName)

    def _getOrigCoord(self):
        return -1. * self.x.get(), -1. * self.y.get(), -1. * self.z.get()

    def _validate(self):
        errors = super(ProtImportVolumes, self)._validate()
        if (not self.filesPattern.empty()) and self.setHalfMaps.get():
            errors.append("You can not use the options 'Pattern' "
                          "and 'Set half maps' simultaneously")

        return errors


class ProtImportPdb(ProtImportFiles):
    """ Protocol to import an atomic structure  to the project.
Format may be PDB or MMCIF"""
    _label = 'import atomic structure'
    IMPORT_FROM_ID = 0
    IMPORT_FROM_FILES = 1
    SYMPLIFIED_STRUCT = "symplified_atom_structure.cif"

    def __init__(self, **args):
        ProtImportFiles.__init__(self, **args)

    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputPdbData', params.EnumParam, choices=['id', 'file'],
                      label="Import atomic structure from",
                      default=self.IMPORT_FROM_ID,
                      display=params.EnumParam.DISPLAY_HLIST,
                      help='Import mmCIF data from online server or local file')
        form.addParam('pdbId', params.StringParam,
                      condition='inputPdbData == IMPORT_FROM_ID',
                      label="Atomic structure ID ", allowsNull=True,
                      help='Type a mmCIF ID (four alphanumeric characters).')
        form.addParam('pdbFile', params.PathParam, label="File path",
                      condition='inputPdbData == IMPORT_FROM_FILES',
                      allowsNull=True,
                      help='Specify a path to desired atomic structure.')
        form.addParam('inputVolume', params.PointerParam, label="Input Volume",
                      pointerClass='Volume',
                      allowsNull=True,
                      help='Associate this volume to the mmCIF file.')

    def _insertAllSteps(self):
        if self.inputPdbData == self.IMPORT_FROM_ID:
            self._insertFunctionStep('pdbDownloadStep')
        else:
            self._insertFunctionStep('createOutputStep', self.pdbFile.get())

    def pdbDownloadStep(self):
        """Download all pdb files in file_list and unzip them.
        """
        aSH = emconv.AtomicStructHandler()
        print("retrieving atomic structure with ID = %s" % self.pdbId.get())
        atomStructPath = aSH.readFromPDBDatabase(self.pdbId.get(),
                                                 type='mmCif',
                                                 dir=self._getExtraPath())
        self.createOutputStep(atomStructPath)

    #        downloadPdb(self.pdbId.get(), pdbPath, self._log)

    def createOutputStep(self, atomStructPath):
        """ Copy the PDB structure and register the output object.
        """
        if not exists(atomStructPath):
            raise Exception("Atomic structure not found at *%s*" % atomStructPath)

        baseName = basename(atomStructPath)
        localPath = abspath(self._getExtraPath(baseName))

        if str(atomStructPath) != str(localPath):  # from local file
            pwutils.copyFile(atomStructPath, localPath)

        localPath = relpath(localPath)

        pdb = emobj.AtomStruct()
        volume = self.inputVolume.get()

        # if a volume exists assign it to the pdb object
        # IMPORTANT: we DO need "if volume is not None"
        # because we need to persist the pdb object
        # before we can make the last source relation
        if volume is not None:
            pdb.setVolume(volume)

        pdb.setFileName(localPath)
        self._defineOutputs(outputPdb=pdb)

        if volume is not None:
            self._defineSourceRelation(volume, pdb)

    def _summary(self):
        if self.inputPdbData == self.IMPORT_FROM_ID:
            summary = ['Atomic structure imported from ID: *%s*' %
                       self.pdbId]
        else:
            summary = ['Atomic structure imported from file: *%s*' %
                       self.pdbFile]

        return summary

    def _validate(self):
        errors = []
        if (self.inputPdbData == self.IMPORT_FROM_FILES and not exists(
                self.pdbFile.get())):
            errors.append("Atomic structure not found at *%s*" %
                          self.pdbFile.get())
        # TODO: maybe also validate that if exists is a valid PDB file
        return errors


class ProtImportSetOfAtomStructs(ProtImportFiles):
    """ Protocol to import a set of atomic structure  to the project.
    Format may be PDB or MMCIF"""
    _label = 'import set of atomic structures'
    _OUTNAME = 'outputAtomStructs'
    _possibleOutputs = {_OUTNAME: emobj.SetOfAtomStructs}

    IMPORT_FROM_ID = 0
    IMPORT_FROM_FILES = 1

    def __init__(self, **args):
      ProtImportFiles.__init__(self, **args)

    def _defineParams(self, form):
      form.addSection(label='Input')
      form.addParam('inputPdbData', params.EnumParam, choices=['id', 'file'],
                    label="Import atomic structures from",
                    default=self.IMPORT_FROM_ID,
                    display=params.EnumParam.DISPLAY_HLIST,
                    help='Import mmCIF data from online server or local files')
      form.addParam('pdbIds', params.StringParam,
                    condition='inputPdbData == IMPORT_FROM_ID',
                    label="Atomic structure IDs ", allowsNull=True,
                    help='Type a mmCIF ID (four alphanumeric characters, comma-separated)\n'
                         'i.e: 5ni1, 1ake')
      form.addParam('filesPath', params.PathParam, label="Files directory path: ",
                    condition='inputPdbData == IMPORT_FROM_FILES',
                    allowsNull=True,
                    help="Specify a path to the directory where the files are stored.\n"
                         "The path can also contain wildcards to select"
                         "from several folders. \n\n"
                         "Examples:\n"
                         "  ~/project/data/day??_files/\n"
                         "Each '?' represents one unknown character\n\n"
                         "  ~/project/data/day*_files/\n"
                         "'*' represents any number of unknown characters\n\n"
                         "  ~/project/data/day##_files/\n"
                         "'##' represents two digits that will be used as "
                         "file ID\n\n"
                         "NOTE: wildcard characters ('*', '?', '#') "
                         "cannot appear in the actual path.)")
      form.addParam('filesPattern', params.StringParam, condition='inputPdbData == IMPORT_FROM_FILES',
                    label='Pattern: ',
                    default="*",
                    help="Pattern of the files to be imported.\n\n"
                         "The pattern can contain standard wildcards such as\n"
                         "*, ?, etc, or special ones like ### to mark some\n"
                         "digits in the filename as ID.\n\n"
                         "NOTE: wildcards and special characters "
                         "('*', '?', '#', ':', '%') cannot appear in the "
                         "actual path.\n\n"
                         "You may create AtomStruct from PDB (.pdb) or CIF/mmCIF (.cif)")

    def _insertAllSteps(self):
      if self.inputPdbData == self.IMPORT_FROM_ID:
        self._insertFunctionStep('pdbDownloadStep')
      else:
        filenames = []
        for fn in glob.glob(join(self.filesPath.get(), self.filesPattern.get())):
          filenames.append(fn)
        self._insertFunctionStep('createOutputStep', filenames)

    def pdbDownloadStep(self):
      """Download all pdb files in file_list and unzip them.
      """
      aSH = emconv.AtomicStructHandler()
      ASPaths = []
      for pdbId in self.pdbIds.get().split(','):
          print("retrieving atomic structure with ID = %s" % pdbId)
          ASPaths.append(aSH.readFromPDBDatabase(pdbId.strip(), type='mmCif', dir=self._getExtraPath()))
      self.createOutputStep(ASPaths)

    def createOutputStep(self, atomStructPaths):
      """ Copy the PDB structures and register the output object.
      """
      outASs = emobj.SetOfAtomStructs().create(self._getPath())
      
      numStructs = len(atomStructPaths)
      padding_format_str = '%0' + str(len(str(numStructs))) + 'd'

      for atomStructPath in atomStructPaths:
          if not exists(atomStructPath):
              raise Exception("Atomic structure not found at *%s*" % atomStructPath)

          baseName = basename(atomStructPath)
          localPath = abspath(self._getExtraPath(baseName))

          if str(atomStructPath) != str(localPath):  # from local file
              if exists(localPath):
                  localPath = splitext(localPath)[0] + padding_format_str % (i+1) + splitext(localPath)[1]
                  i += 1
              else:
                  i = 0
                  
              pwutils.copyFile(atomStructPath, localPath)

          localPath = relpath(localPath)

          outASs.append(emobj.AtomStruct(filename=localPath))

      self._defineOutputs(**{self._OUTNAME:outASs})


    def _summary(self):
      if self.inputPdbData == self.IMPORT_FROM_ID:
        summary = ['Atomic structure imported from IDs: *%s*' %
                   self.pdbIds]
      else:
        summary = ['Atomic structure imported from files: *%s*' %
                   self.filesPattern]

      return summary

    def _validate(self):
      errors = []
      return errors


######################################


def fetch_emdb_map(id, directory, tmpDirectory):
    """ get map from emd
    :param id: 3D MAP ID in EMDB
    :return: local 3Dmap filename
    """
    import socket

    # get computer name and select server
    url_rest_api = "https://www.ebi.ac.uk/emdb/api/entry/map/%s"
    hname = socket.gethostname()
    if hname.endswith('.edu') or hname.endswith('.gov'):
        site = 'ftp.wwpdb.org'
        url_pattern = 'ftp://%s/pub/emdb/structures/EMD-%s/map/%s'
    elif hname.endswith('.cn'):
        site = 'ftp.emdb-china.org'
        url_pattern = 'ftp://%s/structures/EMD-%s/map/%s'
    else:
        site = 'ftp.ebi.ac.uk'
        url_pattern = 'ftp://%s/pub/databases/emdb/structures/EMD-%s/map/%s'

    if len(str(id)) < 4:
        id = '%04d' % id
    else:
        id = '%d' % id

    map_name = 'emd_%s.map' % id
    map_gz_name = map_name + '.gz'
    map_url = url_pattern % (site, id, map_gz_name)
    name = 'EMD-%s' % id
    minimum_map_size = 8192  # bytes
    url_rest_api = url_rest_api % id

    nTimes = 3
    for i in range(nTimes):  # if error repeat the fetch file up to nTimes
        try:
            # raise Exception("test")  # uncomment to test this loop
            map_path, samplingAPI, originAPI = fetch_file(map_url,
                                                          url_rest_api,
                                                          name,
                                                          minimum_map_size,
                                                          directory,
                                                          tmpDirectory,
                                                          map_name
                                                          )
            break
        except Exception as e:
            if i+1<nTimes:
                logger.info("Retrieving 3D map with id=%s failed retrying (%d/%d)" %
                  (id, i+1, nTimes))
            else:
                logger.error("Cannot retrieve File from EMDB", exc_info=e)


    originAPI = array(originAPI) * samplingAPI  # convert to Angstrom
    # check consistency between file header and rest API
    ccp4header = emconv.Ccp4Header(map_path, readHeader=True)
    samplingHeader = ccp4header.computeSampling()  # unit = A/px
    originHeader = array(ccp4header.getOrigin())  # unit = A

    if abs(samplingHeader - samplingAPI) >= 0.01:
        print("###########################\n"
              "WARNING: sampling rate stored in EMDB\n"
              "database and 3D map header file do not match\n"
              "API=%f, header=%f\n"
              "###########################\n" % (samplingAPI, samplingHeader))

    if norm(originHeader - originAPI) >= 0.1:
        print("###########################\n"
              "WARNING: origin  stored in EMDB\n"
              "database and 3D map header file do not match\n"
              "API=%f, header=%f\n"
              "###########################\n" % (originAPI[0], originHeader[0]))
    return map_path, samplingAPI, originAPI


def fetch_file(url, url_rest_api, name,
               minimum_file_size=8192,
               save_dir='',
               tmp_dir='',
               save_name=''):
    """
    Download 3DMAPfile from EMDB

    :param name: EMB name id, format -> EMD-id
    :param url_rest_api: EMDB rest api, here we ay ask for 3D map properties
    :param tmp_dir: directory in which the compressed
                    file will be saved temporarily
    :param url:  3D map url
    :param minimum_file_size:
    :param save_dir: save file in this directory
    :param save_name:  local filename
    :return: local file name
    """
    import urllib.request
    import requests
    noCompressName = join(save_dir, save_name)
    compressName = join(tmp_dir, save_name + ".gz")

    # If uncompressed file exists
    if os.path.exists(noCompressName):
        logger.info("File fetching skip: file %s exists." % noCompressName)
    else:

        logger.info("Fetching file from %s" % url)
        urllib.request.urlretrieve(url, filename=compressName)

    if stat(compressName).st_size < minimum_file_size:
        raise Exception("File Downloaded from EMDB is empty")

    gunzip(compressName, noCompressName)

    logger.info("Fetching metadata from %s" % url_rest_api)
    json_results = requests.get(url_rest_api).json()
    sampling_tag = "pixel_spacing"
    sampling_x_tag = "x"
    origin_tag = "origin"
    origin_x_tag = "col"
    origin_y_tag = "row"
    origin_z_tag = "sec"

    # units A/px
    results = json_results['map']
    sampling = float(results[sampling_tag][sampling_x_tag]['valueOf_'])
    # units unknown may be pixels since this is integer
    x = float(results[origin_tag][origin_x_tag])
    y = float(results[origin_tag][origin_y_tag])
    z = float(results[origin_tag][origin_z_tag])

    return noCompressName, sampling, (x, y, z)


def gunzip(gzpath, path):
    import gzip
    gzf = gzip.open(gzpath)
    f = open(path, 'wb')
    f.write(gzf.read())
    f.close()
    gzf.close()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

"""
Sub-packages containing different sub-modules to import different type
of objects.
"""

from .base import ProtImport, ProtImportFiles
from .coordinates import ProtImportCoordinates, ProtImportCoordinatesPairs
from .ctfs import ProtImportCTF
from .images import ProtImportImages
from .masks import ProtImportMask
from .micrographs import ProtImportMicrographs, ProtImportMovies
from .particles import ProtImportParticles, ProtImportAverages
from .volumes import ProtImportVolumes, ProtImportPdb, ProtImportSetOfAtomStructs
from .sequence import ProtImportSequence



# ***************************************************************************
# * Authors:    Daniel Marchán (da.marchan@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
from pyworkflow.object import Pointer
from pyworkflow.tests import BaseTest, setupTestProject, DataSet
from pwem.protocols.protocol_import import ProtImportMicrographs
from pwem.protocols.protocol_mathematical_operator import ProtMathematicalOperator



class TestMathematicalOperator(BaseTest):
    """ Test mathematical operator protocol"""

    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)

    def testMethods(self):

        prot = self.newProtocol(ProtMathematicalOperator,
                                      boolMain=True)
        self.assertTrue(prot.formulaNeedsParam(ProtMathematicalOperator.PARAM1), "formulaNeedsParam false negative for PARAM1")
        self.assertTrue(prot.formulaNeedsParam(ProtMathematicalOperator.PARAM2), "formulaNeedsParam false negative for PARAM2")

        # validation
        self.assertEqual(0, len(prot.validate()), "validation for default value do not work")

        prot.expression.set("SDSKJSL")
        self.assertFalse(prot.formulaNeedsParam(ProtMathematicalOperator.PARAM1), "formulaNeedsParam false positive for PARAM1")
        self.assertFalse(prot.formulaNeedsParam(ProtMathematicalOperator.PARAM2), "formulaNeedsParam false positive for PARAM2")

        prot.expression.set(ProtMathematicalOperator.PARAM1 + "LOLO")
        self.assertTrue(prot.formulaNeedsParam(ProtMathematicalOperator.PARAM1), "formulaNeedsParam false negative for PARAM1")
        self.assertFalse(prot.formulaNeedsParam(ProtMathematicalOperator.PARAM2), "formulaNeedsParam false positive for PARAM2")


    def testMathematicalOperatorManual(self):

        # Calculate a manual operation
        protNoInput = self.newProtocol(ProtMathematicalOperator)

        protNoInput.expression.set('2 + (3 * 2)')
        protNoInput.setObjLabel("No inputs")
        self.launchProtocol(protNoInput)

        self.assertEqual(protNoInput.result, 8, "Computation without input does not match.")

        # Feed the protocol with inputs
        # Calculate a manual operation
        protWithInput = self.newProtocol(ProtMathematicalOperator)
        pntToOutput = Pointer(protNoInput, extended="result")
        protWithInput.input1.setPointer(pntToOutput)
        protWithInput.expression.set('X1 + 1')
        protWithInput.setObjLabel("X1 as input")
        self.launchProtocol(protWithInput)

        self.assertEqual(protWithInput.result.get(), 9, "Computation without input does not match.")

        # Feed the protocol with X2 input as well
        # Calculate a manual operation
        protWithBothInput = self.newProtocol(ProtMathematicalOperator)
        pntToOutput = Pointer(protNoInput, extended="result")
        protWithBothInput.input1.setPointer(pntToOutput)

        pntToOutput = Pointer(protWithInput, extended="result")
        protWithBothInput.input2.setPointer(pntToOutput)

        protWithBothInput.expression.set('X1 - (X1*X2)')
        protWithBothInput.setObjLabel("X1 and X2 as input")
        self.launchProtocol(protWithBothInput)

        self.assertEqual(protWithBothInput.result.get(), -64, "Computation without input does not match.")




"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from pwem import EM_PROGRAM_ENTRY_POINT, CHIMERA_ENTRY_POINT, \
     CONVERT_ENTRY_POINT
from pwem.constants import URL

from pwem import __version__

from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Load requirements.txt
with open('requirements.txt') as f:
    requirements = f.read().splitlines()


# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='scipion-em',  # Required
    version=__version__,  # Required
    description='This modules contains classes related with EM',  # Required
    long_description=long_description,  # Optional
    url=URL,  # Optional
    author='J.M. De la Rosa Trevin, '
           'Roberto Marabini, '
           'Grigory Sharov, '
           'Josue Gomez Blanco, '
           'Pablo Conesa, '
           'Yunior Fonseca Reyna',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='delarosatrevin@scilifelab.se, '
                 'roberto@cnb.csic.es, '
                 'gsharov@mrc-lmb.cam.ac.uk, '
                 'josue.gomez-blanco@mcgill.ca, '
                 'pconesa@cnb.csic.es, '
                 'fonsecareyna@cnb.csic.es',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        # Pick your license as you wish
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering'
    ],
    keywords='scipion cryoem imageprocessing scipion-3.0',  # Optional
    packages=find_packages(),
    install_requires=[requirements],
    entry_points={
        'console_scripts': [
            '%s = pwem.cmd.program:main' % EM_PROGRAM_ENTRY_POINT,
            '%s = pwem.cmd.chimera_client:main' % CHIMERA_ENTRY_POINT,
            '%s = pwem.cmd.convert:main' % CONVERT_ENTRY_POINT,
        ],
        'pyworkflow.plugin': 'pwem = pwem'
    },
    package_data={
      'pwem': ['scipion_icon.png', 'protocols.conf', 'templates/*', 'cmd/*']
    }
)



# ***************************************************************************
# * Authors:     Airen Zaldivar (azaldivar@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/

import pyworkflow.tests as pwtests

import pwem.protocols as emprot


class TestImportBase(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsXmipp = pwtests.DataSet.getDataSet('xmipp_tutorial')
        # cls.dsRelion = DataSet.getDataSet('relion_tutorial')
        cls.dsGroel = pwtests.DataSet.getDataSet('groel')
        
    def checkOutput(self, prot, outputName, conditions=[]):
        """ Check that an ouput was generated and
        the condition is valid. 
        """
        o = getattr(prot, outputName, None)
        locals()[outputName] = o 
        self.assertIsNotNone(o, "Output: %s is None" % outputName)
        for cond in conditions:
            self.assertTrue(eval(cond), 'Condition failed: ' + cond)
        
    
class TestImportCoordinates(TestImportBase):

    def testImportCoordinates(self):
        # First, import a set of micrographs
        protImport = self.newProtocol(emprot.ProtImportMicrographs,
                                      filesPath=self.dsXmipp.getFile('allMics'),
                                      samplingRate=1.237, voltage=300)
        protImport.setObjLabel('import micrographs from xmipp tutorial ')
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographs.getFileName(),
                             "There was a problem with the import")

        prot1 = self.newProtocol(emprot.ProtImportCoordinates,
                                 importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_XMIPP,
                                 filesPath=self.dsXmipp.getFile('pickingXmipp'),
                                 filesPattern='*.pos', boxSize=550,
                                 scale=3.,
                                 invertX=False,
                                 invertY=False
                                 )
        prot1.inputMicrographs.set(protImport.outputMicrographs)
        prot1.setObjLabel('import coords from xmipp ')
        self.launchProtocol(prot1)
        
        # Make sure that all 264 coordinates where correctly imported
        self.assertTrue(prot1.outputCoordinates.getSize() == 264)

        filepath = prot1.outputCoordinates.getFileName()
        prot2 = self.newProtocol(emprot.ProtImportCoordinates,
                                 importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_SCIPION,
                                 filesPath=filepath, 
                                 filesPattern='',
                                 boxSize=110,
                                 )
        prot2.inputMicrographs.set(protImport.outputMicrographs)
        prot2.setObjLabel('import coords from scipion ')
        self.launchProtocol(prot2)
        # Make sure that all 264 coordinates where correctly imported
        self.assertTrue(prot2.outputCoordinates.getSize() == 264)

        # prot2 = self.newProtocol(ProtImportCoordinates,
        #                          importFrom=ProtImportCoordinates.IMPORT_FROM_RELION,
        #                          filesPath=self.dsXmipp.getFile('boxingDir'),#no dataset with picking
        #                          filesPattern='info/*_info.json',
        #                          boxSize=110,
        #                          scale=2,
        #                          invertX=False,
        #                          invertY=False
        #                          )
        # prot2.inputMicrographs.set(protImport.outputMicrographs)
        # prot2.setObjLabel('import coords from relion ')

        prot3 = self.newProtocol(emprot.ProtImportCoordinates,
                                 importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_EMAN,
                                 filesPath=self.dsXmipp.getFile('boxingDir'),
                                 filesPattern='*_info.json', boxSize=550,
                                 scale=5.,
                                 invertX=False,
                                 invertY=False)
        prot3.inputMicrographs.set(protImport.outputMicrographs)
        prot3.setObjLabel('import coords from eman ')

        self.launchProtocol(prot3)

        # First, import a set of micrographs
        protImportGroel = self.newProtocol(emprot.ProtImportMicrographs,
                                           filesPath=self.dsGroel.getFile('mic1'),
                                           samplingRate=1)
        protImportGroel.setObjLabel('import micrographs from groel')
        self.launchProtocol(protImportGroel)
        self.assertIsNotNone(protImportGroel.outputMicrographs.getFileName(),
                             "There was a problem with the import")

        protPickGroel = self.newProtocol(emprot.ProtImportCoordinates,
                                         importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_DOGPICKER,
                                         filesPath=self.dsGroel.getFile('pickingDogpicker'),
                                         filesPattern='*.txt', boxSize=10,
                                         threshold=0.7)
        protPickGroel.inputMicrographs.set(protImportGroel.outputMicrographs)
        protPickGroel.setObjLabel('import coords from dogpicker ')

        self.launchProtocol(protPickGroel)


class TestImportCoordinatesPairs(TestImportBase):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsRct = pwtests.DataSet.getDataSet('rct')
        cls.micsFn = cls.dsRct.getFile('positions/input_micrographs.xmd')
        cls.patternU1 = cls.dsRct.getFile('positions/F_rct_u_*.pos')
        cls.patternT1 = cls.dsRct.getFile('positions/F_rct_t_*.pos')
        cls.micsUFn1 = cls.dsRct.getFile('untilted')
        cls.micsTFn1 = cls.dsRct.getFile('tilted')

        cls.dsEman = pwtests.DataSet.getDataSet('eman')
        cls.micsUFn2 = cls.dsEman.getFile('micU')
        cls.micsTFn2 = cls.dsEman.getFile('micT')
        cls.patternU2 = cls.dsEman.getFile("coords/ip3r10252011-0005_0-2_info.json")
        cls.patternT2 = cls.dsEman.getFile("coords/ip3r10252011-0005_10_info.json")

    def testImportCoordinatesPairs(self):
        # First, import a set of micrograph pairs
        protImport1 = self.newProtocol(emprot.ProtImportMicrographsTiltPairs,
                                       patternUntilted=self.micsUFn1,
                                       patternTilted=self.micsTFn1,
                                       samplingRate=2.28, voltage=100,
                                       sphericalAberration=2.9)
        protImport1.setObjLabel('import tilt pair micrographs from rct tutorial ')
        self.launchProtocol(protImport1)
        self.assertIsNotNone(protImport1.outputMicrographsTiltPair.getFileName(),
                             "There was a problem with the import")

        protImport2 = self.newProtocol(emprot.ProtImportMicrographsTiltPairs,
                                       patternUntilted=self.micsUFn2,
                                       patternTilted=self.micsTFn2,
                                       samplingRate=2.8, voltage=200,
                                       sphericalAberration=2.0)
        protImport2.setObjLabel('import tilt pair micrographs from eman2 tutorial ')
        self.launchProtocol(protImport2)
        self.assertIsNotNone(protImport2.outputMicrographsTiltPair.getFileName(),
                             "There was a problem with the import")

        prot1 = self.newProtocol(emprot.ProtImportCoordinatesPairs,
                                 importFrom=emprot.ProtImportCoordinatesPairs.IMPORT_FROM_XMIPP,
                                 xmippMdFn=self.micsFn,
                                 patternUntilted=self.patternU1,
                                 patternTilted=self.patternT1,
                                 boxSize=100)
        prot1.inputMicrographsTiltedPair.set(protImport1.outputMicrographsTiltPair)
        prot1.setObjLabel('import coord pairs from xmipp ')
        self.launchProtocol(prot1)

        # Make sure that all 1901 coordinates where correctly imported
        self.assertTrue(prot1.outputCoordinatesTiltPair.getSize() == 1901)

        prot2 = self.newProtocol(emprot.ProtImportCoordinatesPairs,
                                 importFrom=emprot.ProtImportCoordinatesPairs.IMPORT_FROM_EMAN,
                                 patternUntilted=self.patternU2,
                                 patternTilted=self.patternT2,
                                 boxSize=256)
        prot2.inputMicrographsTiltedPair.set(protImport2.outputMicrographsTiltPair)
        prot2.setObjLabel('import coord pairs from eman2 ')
        self.launchProtocol(prot2)

        # Make sure that all 104 coordinates where correctly imported
        self.assertTrue(prot2.outputCoordinatesTiltPair.getSize() == 104)



# ***************************************************************************
# * Authors:    Roberto Marabini (roberto@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
from tempfile import mkstemp

import numpy as np
import os

from pwem.objects import Matrix
from pyworkflow.tests import BaseTest, setupTestProject
from pyworkflow.utils import runJob
from xmipp3 import Plugin
import pwem.protocols as emprot


def createFeatFile(fd):
    # create feat file
    f = os.fdopen(fd, "w")
    f.write("""# XMIPP_STAR_1 *
# Type of feature (sph, blo, gau, Cyl, dcy, cub, ell, con)(Required)
# The operation after adding the feature to the phantom (+/=) (Required)
# The feature density (Required)
# The feature center (Required)
# The vector for special parameters of each vector (Required)
# Sphere: [radius] Blob : [radius alpha m] Gaussian : [sigma]
# Cylinder : [xradius yradius height rot tilt psi]
# DCylinder : [radius height separation rot tilt psi]
# Cube : [xdim ydim zdim rot tilt psi]
# Ellipsoid : [xradius yradius zradius rot tilt psi]
# Cone : [radius height rot tilt psi]
data_block1
 _dimensions3D  '100 100 100'
 _phantomBGDensity  0.
 _scale  1.
data_block2
loop_
 _featureType
 _featureOperation
 _featureDensity
 _featureCenter
 _featureSpecificVector
""")
    theta_max = 8 * np.pi
    nPoints = 100
    thetaStep = theta_max/nPoints/1.2
    for theta in np.arange(-theta_max/2.4, theta_max/2.4, thetaStep):
        z = theta * 3.
        x = np.sin(theta) * theta_max*1.3
        y = np.cos(theta) * theta_max*1.3
        f.write("sph = 1 '%f %f %f' '5'\n" % (x, y, z))
    f.close()


def createParamFile(fd):
    f = os.fdopen(fd, "w")
    f.write("""# XMIPP_STAR_1 *
# Projection Parameters
#_angleFile
# X and Y projection dimensions [Xdim Ydim]
# Rotation range and number of samples [Start Finish NSamples]
# Rotation angle added noise  [noise (bias)]
# Type of range for Rotation (random_group or random or even)
# Tilt range and number of samples for Tilt
# Tilt angle added noise
# Type of range for Tilt
# Psi range and number of samples
# Psi added noise
# Type of range for Psi
# Noise applied to pixels [noise (bias)]
# Noise applied to particle center coordinates [noise (bias)]

data_block1
_dimensions2D   '100 100'
_projRotRange    '0 360 4'
_projRotNoise   '0.0 0.0'
_projRotRandomness   random
_projTiltRange    '0 180 4'
_projTiltNoise   '0.0 0.0'
_projTiltRandomness   random
_projPsiRange    '0 360 4'
_projPsiNoise   '0.0 0.0'
_projPsiRandomness   random
_noisePixelLevel   '0. 0.'
""")
    f.close()


def projectPhantom(featFileName, paramFileName, particlesFileName):
    args = "-i %s -o %s" % (featFileName, particlesFileName)
    args += " --params %s" % paramFileName
    runJob(None, "xmipp_phantom_project", args, env=Plugin.getEnviron())


class TestProtInvertHand(BaseTest):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        (fd, cls.featFileName) = mkstemp(suffix=".feat")
        createFeatFile(fd)
        (fd, cls.paramFileName) = mkstemp(suffix=".param")
        createParamFile(fd)
        (fd, cls.stackFileName) = mkstemp(suffix=".stk")
        cls.xmdFileName = cls.stackFileName.replace(".stk", ".xmd")
        projectPhantom(cls.featFileName, cls.paramFileName,
                       cls.stackFileName)
        (fd, cls.stackFileName2) = mkstemp(suffix=".stk")
        cls.xmdFileName2 = cls.stackFileName2.replace(".stk", ".xmd")
        projectPhantom(cls.featFileName, cls.paramFileName,
                       cls.stackFileName2)

    def test_01_autoaligment(self):
        # import first set of particles
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_XMIPP3,
                                 mdFile=self.xmdFileName,
                                 magnification=10000,
                                 samplingRate=1,
                                 haveDataBeenPhaseFlipped=False
                                 )
        prot1.setObjLabel('import Particles 1')
        self.launchProtocol(prot1)

        # assign transformation from 1 set to second set
        prot3 = self.newProtocol(emprot.ProtAlignmentInvertHand,
                                 inputParticles=prot1.outputParticles,
                                 )
        prot3.setObjLabel('assign angle 1 -> 2')
        self.launchProtocol(prot3)
        for part1, part3 in emprot.izip(
                prot1.outputParticles,
                prot3.outputParticles):
            matrix = Matrix()
            matrix.copy(part1.getTransform())
            m = matrix.getMatrix()
            m[0, 2] *= -1.
            m[1, 2] *= -1.
            m[2, 1] *= -1.
            m[2, 0] *= -1.
            result = np.allclose(m,
                                 part3.getTransform().getMatrix())
            self.assertTrue(result)
            result = np.allclose(part1.getTransform().getMatrix(),
                                 part3.getTransform().getMatrix())
            self.assertFalse(result)
        self.assertEqual(len(prot3.outputParticles),
                         len(prot1.outputParticles))



# ***************************************************************************
# * Authors:     Mohsen Kazemi (mkazemi@cnb.csic.es)
# *              Javier Vargas (javier.vargasbalbuena@mcgill.ca)  
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/

import os

from pyworkflow.tests import BaseTest, setupTestProject, DataSet

from pwem import Domain
from pwem.objects.data import SetOfParticles
from pwem.protocols import ProtImportParticles, ProtImportVolumes, ProtSubSet


class TestVolumeHomogenizer(BaseTest):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        cls.dsRelion = DataSet.getDataSet('relion_tutorial')
        cls.dsXmipp = DataSet.getDataSet('xmipp_tutorial')

    def importFromRelionRefine3D(self):
        """ Import aligned Particles
        """
        prot = self.newProtocol(ProtImportParticles,
                                objLabel='particles from relion (auto-refine 3d)',
                                importFrom=ProtImportParticles.IMPORT_FROM_RELION,
                                starFile=self.dsRelion.getFile('import/classify3d/extra/relion_it015_data.star'),
                                magnification=10000,
                                samplingRate=7.08,
                                haveDataBeenPhaseFlipped=True
                                )
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputParticles.getFileName(),
                             "There was a problem with the import")

        return prot

    def getParticles(self, protImport, classid):
        dbPartSet = protImport._getPath("particles_class-%d.sqlite" % classid)
        class3D = protImport.outputClasses[classid]
        if os.path.exists(dbPartSet):
            os.remove(dbPartSet)
        partSet = SetOfParticles(filename=dbPartSet)
        partSet.copyInfo(class3D)

        for part in class3D:
            partSet.append(part)
        partSet.write()
        partSet.close()

        protImportCls1 = self.newProtocol(ProtImportParticles,
                                          objLabel='particles class-%d' % classid,
                                          importFrom=ProtImportParticles.IMPORT_FROM_SCIPION,
                                          sqliteFile=dbPartSet,
                                          magnification=10000,
                                          samplingRate=7.08,
                                          haveDataBeenPhaseFlipped=True
                                          )
        self.launchProtocol(protImportCls1)
        self.assertIsNotNone(protImportCls1.outputParticles.getFileName(),
                             "There was a problem with the import")
        return protImportCls1

    def subSetOfParticles(self, importPartClass1):

        protSubset = self.newProtocol(ProtSubSet,
                                      objLabel='subset of particles',
                                      chooseAtRandom=True,
                                      nElements=200)

        protSubset.inputFullSet.set(importPartClass1.outputParticles)
        self.launchProtocol(protSubset)
        self.assertIsNotNone(protSubset.outputParticles.getFileName(),
                             "There was a problem with the protSubset")

        return protSubset

    def getVolume(self, classid):
        """ Import a Volumes
        """
        protImportVol = self.newProtocol(ProtImportVolumes,
                                         objLabel='vol class-%d' % classid,
                                         filesPath=self.dsRelion.getFile(
                                             'import/classify3d/extra/relion_it015_class00%d.mrc' %
                                             classid),
                                         samplingRate=7.08,
                                         )
        self.launchProtocol(protImportVol)
        return protImportVol

    def relionRefine(self, protSubSetClass, vol, classid):

        ProtRelionRefine3D = Domain.importFromPlugin('relion.protocols',
                                                     'ProtRelionRefine3D',
                                                     doRaise=True)
        relionRefine = self.newProtocol(ProtRelionRefine3D,
                                        objLabel='relion_%s' % classid,
                                        doCTF=False, runMode=1,
                                        memoryPreThreads=1,
                                        maskDiameterA=424,
                                        angularSamplingDeg=1,
                                        localSearchAutoSamplingDeg=2,
                                        symmetryGroup="c1",
                                        numberOfMpi=3, numberOfThreads=1)

        relionRefine.inputParticles.set(protSubSetClass.outputParticles)
        relionRefine.referenceVolume.set(vol.outputVolume)

        self.launchProtocol(relionRefine)
        return relionRefine

    def homoNoGoldStandardNoAlign(self, relion1, relion2):
        """test without Goldstandard and alignment"""
        XmippProtVolumeHomogenizer = Domain.importFromPlugin('xmipp3.protocols',
                                                             'XmippProtEnrich',
                                                             doRaise=True)
        protVolumeHomogenizer = self.newProtocol(XmippProtVolumeHomogenizer,
                                                 objLabel='volume homogenizer1')

        protVolumeHomogenizer.referenceVolume.set(relion1.outputVolume)
        protVolumeHomogenizer.inputVolume.set(relion2.outputVolume)
        protVolumeHomogenizer.inputParticles.set(relion2.outputParticles)
        protVolumeHomogenizer.doAlignment.set(False)
        self.launchProtocol(protVolumeHomogenizer)

        self.assertIsNotNone(protVolumeHomogenizer.outputParticles.getFileName(),
                             "There was a problem with the homoNoGoldStandardNoAlign")

        return protVolumeHomogenizer

    def homoNoGoldStandardAlign(self, relion1, relion2):
        """test without Goldstandard and alignment"""
        XmippProtVolumeHomogenizer = Domain.importFromPlugin('xmipp3.protocols',
                                                             'XmippProtEnrich',
                                                             doRaise=True)
        protVolumeHomogenizer = self.newProtocol(XmippProtVolumeHomogenizer,
                                                 objLabel='volume homogenizer2')

        protVolumeHomogenizer.referenceVolume.set(relion1.outputVolume)
        protVolumeHomogenizer.inputVolume.set(relion2.outputVolume)
        protVolumeHomogenizer.inputParticles.set(relion2.outputParticles)
        protVolumeHomogenizer.doAlignment.set(True)
        self.launchProtocol(protVolumeHomogenizer)

        self.assertIsNotNone(protVolumeHomogenizer.outputParticles.getFileName(),
                             "There was a problem with the homoNoGoldStandardAlign")

        return protVolumeHomogenizer

    def homoGoldStandardNoAlign(self, relion1, relion2):
        """test without Goldstandard and alignment"""
        XmippProtVolumeHomogenizer = Domain.importFromPlugin('xmipp3.protocols',
                                                             'XmippProtEnrich',
                                                             doRaise=True)
        protVolumeHomogenizer = self.newProtocol(XmippProtVolumeHomogenizer,
                                                 objLabel='volume homogenizer3',
                                                 doGoldStandard=True)

        protVolumeHomogenizer.referenceVolume1.set(relion1.outputVolume)
        protVolumeHomogenizer.inputVolume1.set(relion2.outputVolume)
        protVolumeHomogenizer.referenceVolume2.set(relion1.outputVolume)
        protVolumeHomogenizer.inputVolume2.set(relion2.outputVolume)
        protVolumeHomogenizer.inputParticles.set(relion2.outputParticles)
        protVolumeHomogenizer.doAlignment.set(False)

        self.launchProtocol(protVolumeHomogenizer)
        self.assertIsNotNone(protVolumeHomogenizer.outputParticles01.getFileName(),
                             "There was a problem with the homoGoldStandardNoAlign")

        self.assertIsNotNone(protVolumeHomogenizer.outputParticles02.getFileName(),
                             "There was a problem with the homoGoldStandardNoAlign")

        return protVolumeHomogenizer

    def homoGoldStandardAlign(self, relion1, relion2):
        """test without Goldstandard with alignment"""
        XmippProtVolumeHomogenizer = Domain.importFromPlugin('xmipp3.protocols',
                                                             'XmippProtEnrich',
                                                             doRaise=True)
        protVolumeHomogenizer = self.newProtocol(XmippProtVolumeHomogenizer,
                                                 objLabel='volume homogenizer4',
                                                 doGoldStandard=True)

        protVolumeHomogenizer.referenceVolume1.set(relion1.outputVolume)
        protVolumeHomogenizer.inputVolume1.set(relion2.outputVolume)
        protVolumeHomogenizer.referenceVolume2.set(relion1.outputVolume)
        protVolumeHomogenizer.inputVolume2.set(relion2.outputVolume)
        protVolumeHomogenizer.inputParticles.set(relion2.outputParticles)
        protVolumeHomogenizer.doAlignment.set(True)

        self.launchProtocol(protVolumeHomogenizer)
        self.assertIsNotNone(protVolumeHomogenizer.outputParticles01.getFileName(),
                             "There was a problem with the homoGoldStandardAlign")

        self.assertIsNotNone(protVolumeHomogenizer.outputParticles02.getFileName(),
                             "There was a problem with the homoGoldStandardAlign")

        return protVolumeHomogenizer

    def test_volumeHomogenizer(self):
        """ Test protocol volume homogenizer
        """
        protImpClasses = self.importFromRelionRefine3D()

        importPartClass1 = self.getParticles(protImpClasses, classid=1)
        protSubSetClass1 = self.subSetOfParticles(importPartClass1)

        importPartClass2 = self.getParticles(protImpClasses, classid=2)
        protSubSetClass2 = self.subSetOfParticles(importPartClass2)

        vol1 = self.getVolume(classid=1)
        vol2 = self.getVolume(classid=2)

        relion1 = self.relionRefine(protSubSetClass1, vol1, classid=1)
        relion2 = self.relionRefine(protSubSetClass2, vol2, classid=2)

        homo1 = self.homoNoGoldStandardNoAlign(relion1, relion2)
        # homo2 = self.homoNoGoldStandardAlign(relion1,relion2)
        homo3 = self.homoGoldStandardNoAlign(relion1, relion2)
        # homo4 = self.homoGoldStandardAlign(relion1,relion2)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import pyworkflow.protocol.params as params

import pwem.objects as emobj
from pwem.protocols import ProtCTFMicrographs


class ProtCTFAssign(ProtCTFMicrographs):
    """ This protocol assigns a CTF estimation to a particular
    set of particles producing a new set. """
    _label = 'assign ctf'
    _unionTypes = ['Micrographs',
                   'Particles']
    
    def __init__(self, **kwargs):
        ProtCTFMicrographs.__init__(self, **kwargs)
        # We need to trace the changes of 'inputType' to 
        # dynamically modify the property of pointerClass
        # of the 'inputSets' parameter

        def onChangeInputType():
            pointerClass = 'SetOf' + self.getEnumText('inputType')
            self.inputSetsParam.setPointerClass(pointerClass)
        # Initial update
        onChangeInputType()
        # Now keep track of changes and update
        self.inputType.trace(onChangeInputType)
    
    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputType', params.EnumParam, choices=self._unionTypes,
                      default=0,  # Micrographs
                      label='Input type:',
                      help='Select the type of objects that you want to assign'
                           ' the CTF.')
        self.inputSetsParam = form.addParam('inputSet', params.PointerParam,
                                            pointerClass='EMSet',
                                            label='Input set',
                                            help='Select the images '
                                                 '(micrographs or particles) '
                                                 'that you want to update the '
                                                 'CTF parameters.')        
        form.addParam('inputCTF', params.PointerParam,
                      pointerClass='SetOfCTF, SetOfParticles',
                      label="Input CTF",
                      help='Select the CTFs that will be used to update '
                           'particles. It can be another set of particles')
        form.addParallelSection(threads=0, mpi=0) 
    
    # --------------------------- INSERT steps functions ----------------------
                                
    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep')
    
    # --------------------------- STEPS functions -----------------------------
    def createOutputStep(self):
        inputSet = self.inputSet.get()
        inputCTF = self.inputCTF.get()
        
        if isinstance(inputSet, emobj.SetOfParticles):
            self._particlesOutputStep(inputSet, inputCTF)
        else:
            self._micrographsOutputStep(inputSet, inputCTF)
            
    def _particlesOutputStep(self, inputSet, inputCTF):
        outputParts = self._createSetOfParticles()
        outputParts.copyInfo(inputSet)
        outputParts.setHasCTF(True)

        if isinstance(inputCTF, emobj.SetOfParticles):
            outputParts.copyItems(inputSet,
                                  itemDataIterator=inputCTF.iterItems(),
                                  updateItemCallback=self._updateItem)
        else:
            ctfDict = {}
            firstCoord = inputSet.getFirstItem().getCoordinate()
            hasMicName = firstCoord.getMicName() is not None

            for ctf in inputCTF:
                if hasMicName:
                    ctfName = ctf.getMicrograph().getMicName()
                else:
                    ctfName = ctf.getMicrograph().getObjId()
    #             print("ctf: ", ctf.printAll(), ctfName)
                ctfDict[ctfName] = ctf.clone()

            missingSet = set()  # Report missing micrographs only once

            for particle in inputSet:
                if particle.hasCoordinate():
                    coord = particle.getCoordinate()
                    micKey = coord.getMicName() if hasMicName else particle.getMicId()
                else:
                    micKey = particle.getMicId()

                if micKey not in missingSet:
                    ctf = ctfDict.get(micKey, None)

                    if ctf is None:
                        self.warning("Discarding particles from micrograph with"
                                     " micName: %s, CTF not found. " % micKey)
                        missingSet.add(micKey)
                    else:
                        newParticle = particle.clone()
                        newParticle.setCTF(ctf)
                        outputParts.append(newParticle)
        
        self._defineOutputs(outputParticles=outputParts)
        self._defineSourceRelation(self.inputSet, outputParts)
        self._defineSourceRelation(self.inputCTF, outputParts)

    def __findCTF(self, inputSet, outputSet, ctfDict, keyFunc):
        for mic in inputSet:
            micKey = keyFunc(mic)
            ctf = ctfDict.get(micKey, None)
            if ctf is None:
                self.warning("Discarding micrographs with micName: %s, CTF not found. " % micKey)
            else:
                newMic = mic.clone()
                outputSet.append(newMic)
        
    def _micrographsOutputStep(self, inputSet, inputCTF):
        outputMics = self._createSetOfMicrographs()
        outputMics.copyInfo(inputSet)
        ctfDict = {}
        
        for ctf in inputCTF:
            mic = ctf.getMicrograph()
            ctfDict[mic.getMicName()] = ctf
            ctfDict[mic.getObjId()] = ctf
        
        # Try first to find the ctf by the micrograph micName
        self.__findCTF(inputSet, outputMics, ctfDict, lambda mic: mic.getMicName())
        
        # Now, if no ctf was found, try using the micId
        if outputMics.getSize() == 0:
            self.warning("No CTF found using micName, now trying with micId")
            self.__findCTF(inputSet, outputMics, ctfDict, lambda mic: mic.getObjId())            
        
        self._defineOutputs(outputMicrographs=outputMics)
        self._defineSourceRelation(self.inputSet, outputMics)
        self._defineCtfRelation(outputMics, self.inputCTF)
    
    # --------------------------- INFO functions ------------------------------
    def _summary(self):
        summary = []
        return summary    
    
    def _methods(self):
        return []
    
    def _validate(self):
        """ The function of this hook is to add some validation before the
        protocol is launched to be executed. It should return a list of errors.
        If the list is empty the protocol can be executed.
        """
        errors = []
        # Add some errors if input is not valid
        inputSet = self.inputSet.get()
        if isinstance(inputSet, emobj.SetOfParticles):
            part = inputSet.getFirstItem()
            if not part.hasMicId():
                errors.append("The input particles doesn't have any micrograph"
                              " assigned.")

        # same micrographs in both CTF??
        return errors
    
    # --------------------------- UTILS functions -----------------------------
    def _stepsCheck(self):
        pass

    def _updateItem(self, item, row):
        item.setCTF(row.getCTF())



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              David Maluenda (dmaluenda@cnb.csic.es)  -streaming version-
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import numpy as np
import time
from datetime import datetime

import pyworkflow.protocol.params as params
from pyworkflow.protocol.constants import STATUS_NEW
import pyworkflow.utils as pwutils
from pyworkflow.object import Float

from pwem.protocols import ProtParticlePickingAuto
import pwem.constants as emcts
import pwem.objects as emobj


class ProtExtractCoords(ProtParticlePickingAuto):
    """ 
    Extract the coordinates information from a set of particles.
    
    This protocol is useful when we want to re-extract the particles
    (maybe resulting from classification and cleaning) with the 
    original dimensions. It can be also handy to visualize the resulting
    particles in their location on micrographs.
    """

    _label = 'extract coordinates'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inputParticles', params.PointerParam,
                      pointerClass='SetOfParticles',
                      label='Input particles', important=True,
                      help='Select the particles from which you want\n'
                           'to extract the coordinates and micrographs.')

        form.addParam('inputMicrographs', params.PointerParam,
                      pointerClass='SetOfMicrographs',
                      label='Input micrographs', important=True,
                      help='Select the micrographs to which you want to\n'
                           'associate the coordinates from the particles.')

        form.addParam('applyShifts', params.BooleanParam, default=False,
                      label='Apply particle shifts?',
                      help='Apply particle shifts (ONLY INTEGER PART) from 2D alignment to '
                           'recalculate new coordinates. This can be useful '
                           'for re-centering particle coordinates.\n'
                           'IMPORTANT: Only the integer part of the shifts will be applied in '
                           'order to avoid interpolation. If you are re-extracting particles and '
                           'want to apply the remaining decimal part of the shifts, set to  "yes" the '
                           'option "Were particle shifts applied?" in alignment assign protocol.'
                      )

        form.addParallelSection(threads=0, mpi=0)

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self.streamingModeOn = self.getInputParticles().isStreamOpen()

        if self.streamingModeOn:
            self.inputSize = 0
            self.outputSize = 0
            self.micsDone = []

            t0 = time.time()
            newParts, self.streamClosed = self.loadInputs()
            print("loadInputs() time: %fs" % (time.time() - t0))

            stepsIds = self._insertNewSteps(newParts)
            self._insertFunctionStep('createOutputStep',
                                     prerequisites=stepsIds, wait=True)
        else:
            self._insertFunctionStep('createOutputStep')

    def _insertNewSteps(self, partIds):
        deps = []
        stepId = self._insertFunctionStep('extractCoordsStep', partIds,
                                          prerequisites=[])
        deps.append(stepId)
        return deps

    def _stepsCheck(self):
        # To allow streaming picking we need to detect:
        #   1) new micrographs ready to be picked
        #   2) new output coordinates that have been produced and add then
        #      to the output set.
        if self.streamingModeOn:
            self._checkNewInput()
            self._checkNewOutput()
        else:
            pass

    def _checkNewInput(self):
        # Check if there are new particles to process from the input set
        partsFile = self.getInputParticles().getFileName()
        micsFile = self.getInputMicrographs().getFileName()
        now = datetime.now()
        self.lastCheck = getattr(self, 'lastCheck', now)
        mTimeParts = datetime.fromtimestamp(os.path.getmtime(partsFile))
        mTimeMics = datetime.fromtimestamp(os.path.getmtime(micsFile))
        # If the input movies.sqlite have not changed since our last check,
        # it does not make sense to check for new input data
        if self.lastCheck > mTimeParts and self.lastCheck > mTimeMics:
            return None
        self.lastCheck = now

        newParts, self.streamClosed = self.loadInputs()

        if len(newParts) > 0:
            fDeps = self._insertNewSteps(newParts)
            outputStep = self._getFirstJoinStep()
            if outputStep is not None:
                outputStep.addPrerequisites(*fDeps)
            self.updateSteps()

    def extractCoordsStep(self, partsIds):
        outputCoords = self.extractCoordinates(partsIds)

        self.outputSize += len(outputCoords)
        t0 = time.time()
        outputCoords.write()
        outputCoords.close()
        print("write time: %fs" % (time.time() - t0))

    def extractCoordinates(self, partsIds=None):
        inPart = self.getInputParticles()
        inMics = self.getInputMicrographs()
        scale = inPart.getSamplingRate() / inMics.getSamplingRate()
        print("Scaling coordinates by a factor *%0.2f*" % scale)
        alignType = inPart.getAlignment()

        suffix = self.getSuffix(partsIds[0]) if partsIds is not None else ''
        #outputCoords = self._createSetOfCoordinates(inMics, suffix=suffix)
        outputCoords = self._createSetOfCoordinates(self.getInputMicrographsPointer(), suffix=suffix)
        # Prepare a double key dictionary to do the match: micname and micId
        micDict = dict()
        for mic in inMics.iterItems():
            # Clone the mics! otherwise we will get pointers and
            # will end up with the same mic in the dictionary.
            clonedMic = mic.clone()
            micDict[clonedMic.getObjId()] = clonedMic
            micDict[clonedMic.getMicName()] = clonedMic

        def appendCoordFromParticle(part):
            coord = part.getCoordinate()

            # Try micname
            micName = coord.getMicName()
            mic = micDict.get(micName, None)

            # Try micid
            if mic is None:
                micKey = coord.getMicId()
                mic = micDict.get(micKey, None)

            if mic is None:
                print("Skipping particle, %s or id %s not found" % (micName, micKey))
            else:
                newCoord.copyObjId(part)
                x, y = coord.getPosition()

                if self.applyShifts:
                    # Get the shifts, they are returned with the sign reverted
                    shifts = self.getShifts(part.getTransform(), alignType)

                    # Add the shifts (values are inverted so subtract)
                    x -= shifts[0]
                    y -= shifts[1]

                # Apply the scale
                x *= scale
                y *= scale

                # Round coordinates to closer integer 39.9 --> 40 and not 39
                finalX = round(x)
                finalY = round(y)

                # Annotate fractions if shifts applied
                if self.applyShifts:
                    newCoord.xFrac = Float(finalX - x)
                    newCoord.yFrac = Float(finalY-y)
                newCoord.setPosition(finalX, finalY)

                newCoord.setMicrograph(mic)
                outputCoords.append(newCoord)

        newCoord = emobj.Coordinate()
        if self.streamingModeOn:
            for partId in partsIds:
                particle = inPart[partId]
                appendCoordFromParticle(particle)
        else:
            for particle in inPart:
                appendCoordFromParticle(particle)

        boxSize = inPart.getXDim() * scale
        outputCoords.setBoxSize(boxSize)

        return outputCoords

    def _checkNewOutput(self):
        if getattr(self, 'finished', False):
            return

        self.finished = self.streamClosed and \
                        self.inputSize == self.outputSize

        streamMode = emobj.Set.STREAM_CLOSED if self.finished else emobj.Set.STREAM_OPEN

        # we will read all ready files
        files = pwutils.glob(self.getTmpOutputPath('*'))
        newData = len(files) > 0
        lastToClose = self.finished and hasattr(self, 'outputCoordinates')
        if newData or lastToClose:
            outSet = self._loadOutputSet()
            if newData:
                for tmpFile in files:
                    tmpSet = emobj.SetOfCoordinates(filename=tmpFile)
                    tmpSet.loadAllProperties()
                    outSet.copyItems(tmpSet)
                    outSet.setBoxSize(tmpSet.getBoxSize())

                    tmpSet.close()
                    pwutils.cleanPath(tmpFile)

            self._updateOutputSet('outputCoordinates', outSet, state=streamMode)

        if self.finished:  # Unlock createOutputStep if finished all jobs
            outputStep = self._getFirstJoinStep()
            if outputStep and outputStep.isWaiting():
                outputStep.setStatus(STATUS_NEW)

    def _loadOutputSet(self):
        setFile = self._getPath("coordinates.sqlite")
        if os.path.exists(setFile):
            outputSet = emobj.SetOfCoordinates(filename=setFile)
            outputSet.loadAllProperties()
            outputSet.enableAppend()
        else:
            outputSet = emobj.SetOfCoordinates(filename=setFile)
            outputSet.setStreamState(outputSet.STREAM_OPEN)
            self._store(outputSet)
            self._defineTransformRelation(self.getInputParticles(), outputSet)
            self._defineSourceRelation(self.getInputMicrographs(), outputSet)

        outputSet.setMicrographs(self.getInputMicrographsPointer())

        return outputSet

    def _getFirstJoinStep(self):
        for s in self._steps:
            if s.funcName == self._getFirstJoinStepName():
                return s
        return None

    def _getFirstJoinStepName(self):
        # This function will be used for streaming, to check which is
        # the first function that need to wait for all micrographs
        # to have completed, this can be overwritten in subclasses
        # (e.g., in Xmipp 'sortPSDStep')
        return 'createOutputStep'

    def createOutputStep(self):
        if not self.streamingModeOn:
            outputCoords = self.extractCoordinates()
            self._defineOutputs(outputCoordinates=outputCoords)
            self._defineSourceRelation(self.inputParticles, outputCoords)
            self._defineSourceRelation(self.inputMicrographs, outputCoords)

    # ------------- UTILS functions ----------------
    def getSuffix(self, suffix):
        return "_tmp%s" % suffix

    def getTmpOutputPath(self, suffix):
        return self._getPath("coordinates%s.sqlite" % self.getSuffix(suffix))

    def loadInputs(self):
        micsFn = self.getInputMicrographs().getFileName()
        micsSet = emobj.SetOfMicrographs(filename=micsFn)
        micsSet.loadAllProperties()

        availableMics = []
        for mic in micsSet:
            availableMics.append(mic.getObjId())

        micsSetClosed = micsSet.isStreamClosed()
        micsSet.close()

        partsFn = self.getInputParticles().getFileName()
        partsSet = emobj.SetOfParticles(filename=partsFn)
        partsSet.loadAllProperties()

        newParts = []
        newMics = []
        for item in partsSet:
            micKey = item.getCoordinate().getMicId()
            if micKey not in self.micsDone and micKey in availableMics:
                newParts.append(item.getObjId())
                if micKey not in self.micsDone:
                    newMics.append(micKey)
        self.micsDone += newMics
        self.inputSize = partsSet.getSize()
        partSetClosed = partsSet.isStreamClosed()
        partsSet.close()

        return newParts, micsSetClosed and partSetClosed

    def getShifts(self, transform, alignType):
        """
        is2D == True-> matrix is 2D (2D images alignment)
                otherwise matrix is 3D (3D volume alignment or projection)
        invTransform == True  -> for xmipp implies projection
                              -> for xmipp implies alignment
        """
        if alignType == emcts.ALIGN_NONE:
            return None

        inverseTransform = alignType == emcts.ALIGN_PROJ
        # only flip is meaningful if 2D case
        # in that case the 2x2 determinant is negative
        flip = False
        matrix = transform.getMatrix()
        if alignType == emcts.ALIGN_2D:
            # get 2x2 matrix and check if negative
            flip = bool(np.linalg.det(matrix[0:2, 0:2]) < 0)
            if flip:
                matrix[0, :2] *= -1.  # invert only the first two columns keep x
                matrix[2, 2] = 1.  # set 3D rot

        elif alignType == emcts.ALIGN_3D:
            flip = bool(np.linalg.det(matrix[0:3, 0:3]) < 0)
            if flip:
                matrix[0, :4] *= -1.  # now, invert first line including x
                matrix[3, 3] = 1.  # set 3D rot

            # flip = bool(numpy.linalg.det(matrix[0:3,0:3]) < 0)
            # if flip:
            #    matrix[0,:4] *= -1.#now, invert first line including x
        shifts = self.geometryFromMatrix(matrix, inverseTransform)

        return shifts

    def geometryFromMatrix(self, matrix, inverseTransform):
        from pwem.convert.transformations import translation_from_matrix
        if inverseTransform:
            matrix = np.linalg.inv(matrix)
            shifts = -translation_from_matrix(matrix)
        else:
            shifts = translation_from_matrix(matrix)
        return shifts

    def getInputParticles(self):
        return self.inputParticles.get()

    # --------------------------- INFO functions ------------------------------
    def _summary(self):
        summary = []
        ps1 = self.getInputParticles().getSamplingRate()
        ps2 = self.getInputMicrographs().getSamplingRate()
        summary.append(u'Input particles pixel size: *%0.3f* (Å/px)' % ps1)
        summary.append(u'Input micrographs pixel size: *%0.3f* (Å/px)' % ps2)
        summary.append('Scaling coordinates by a factor of *%0.3f*' % (ps1 / ps2))
        if self.applyShifts:
            summary.append('Applied 2D shifts from particles')

        if hasattr(self, 'outputCoordinates'):
            summary.append('Output coordinates: *%d*'
                           % self.outputCoordinates.getSize())

        return summary

    def _methods(self):
        # No much to add to summary information
        return self._summary()

    def _validate(self):
        """ The function of this hook is to add some validation before the
        protocol is launched to be executed. It should return a list of errors.
        If the list is empty the protocol can be executed.
        """
        errors = []
        inputParticles = self.getInputParticles()
        first = inputParticles.getFirstItem()
        if first.getCoordinate() is None:
            errors.append('The input particles do not have coordinates!!!')

        if self.applyShifts and not inputParticles.hasAlignment():
            errors.append('Input particles do not have alignment information!')

        return errors
